/**********************************************************************
 * File	: carsmanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 12/07/2007
 *********************************************************************/

#ifndef CARSMANAGER_HPP_INC_19_03_2009___41607409_
#define CARSMANAGER_HPP_INC_19_03_2009___41607409_ 1

#include <list>
#include "geometry.hpp"
#include "spritemanager.hpp"
#include "timer.hpp"
#include "circularlist.hpp"
#include "gunner.hpp"

#define SPEED_MULTIPLIER 50

using std::list;

class TowerInstance;

// une voiture est aussi un tireur.
struct Car : Gunner
{
	enum DIRECTION_en
	{
		DIR_STRAIGHT,
		DIR_RIGHT,
		DIR_LEFT
	};

private:

	float GetWheelFactor(void)
	{
		switch (wheel)
		{
			case DIR_RIGHT:
				return -1.0f;
			case DIR_LEFT:
				return +1.0f;
			default:
				return 0.0f;
		}
	}

	DIRECTION_en wheel;
	Vector2f posf;

	typedef list< TowerInstance* > TwrList;
	typedef TwrList::iterator      TwrListIt;

	TwrList trackers;	  // liste des tourelles qui nous ont pris pour cible (permet de leur signaler notre destruction)

	inline void PartialCopyInto(Car* dest);
	
	int RaiseScore(int points)
	{
		return score_cumul += Gunner::RaiseScore(points);
	}

public:

	//	 === variables publiques ===

	//Rect rect;			  // dans la classe de base "Gunner"
	//Vector2f direction;
	//float angle;

	// -- game action --

	float speed;
	int	health;
	int shield;
	int	score_cumul;

	//	 === methodes ===
	
public:
	
	void InflictDamage(int dmg)
	{
		shield -= dmg;
		if (shield < 0)
		{
			health -= -shield;
			shield = 0;
		}
	}

	// -- update --

	void Update(float delta_time);

	// -- geometrie --

	void SetDirection(DIRECTION_en edir)
	{
		wheel = edir;
	}
	void SetPosition(int x, int y)
	{
		posf.x = x;
		posf.y = y;
		rect.pos.x = x;
		rect.pos.y = y;
	}

	ORect GetORect(void) const
	{
		ORect orc = Rect(0, 0, 40, 24);
		orc.rect.SetCenter(rect.GetCenter());
		orc.dir = direction;
		return orc;
	}

	virtual Vector2i GetSideOffsets(void) const
	{
		return Vector2i(14, 22);
	}
	
	Vector2f GetRawPosf(void)
	{
		return posf;
	}

	void RegisterTracker(TowerInstance* ti);
	void UnregisterTracker(TowerInstance* ti);

	// -- construction / destruction --

	Car(void)
			: wheel(DIR_STRAIGHT),
			  posf(0, 0),
			  speed(1.5f),
			  health(100),
			  shield(0),
			  score_cumul(0) { }

	~Car();
};


// ---------------------------
// **********
// ---------------------------

class CarsManager
{
protected:

	struct ManagedCar
	{
		Car car;
		int spriteId;
	};
	struct PosAndTime
	{
		Vector2i pos;
		float stamp;

		PosAndTime(void) { }

		PosAndTime(Vector2i const& ap, float as)
				: pos(ap),
				  stamp(as) { }
	};

	// manager de sprites :
	SpriteManager sprites;

	// liste des voitures gerees :
	typedef list< ManagedCar >::iterator CarIT;
	list< ManagedCar > cars;

	// liste des épaves :
	CircularList< Vector2i > wrecks;

	// liste des explosions en cours :
	CircularList< PosAndTime > bombs;

	typedef map< const char*, int >::iterator MapIT;
	typedef map< const char*, int >::value_type MapVType;
	map< const char*, int > fn2idMap;

	double time;
	float prev_time;

	int wreckId, explodeId;
	
	struct CarPositionOverviewInfos
	{
		int redPixelId, yellowPixelId;
		SpriteManager* pnoscroll_spm;
		Vector2i level_size;
		CarPositionOverviewInfos() : redPixelId(-1),
									 yellowPixelId(-1),
									 pnoscroll_spm(0),
									 level_size(0, 0)
		{}
	} pos_view;

	static int instances;

public:

	// structure a renvoyer a l'utilisateur, il peut acceder aux voitures et demander une suppression
	// en temps constant grace à ca.
	struct CarAcsor  // Car Accessor.  (sorte d'iterateur non iterable)
	{
		friend class CarsManager;

		bool IsValid(void) const
		{
			return (pValid.GetPtr() && *pValid == true);
		}

		Car& operator * (void)
		{
			BOOST_ASSERT(IsValid());
			return it->car;
		}
		Car* operator -> (void)
		{
			BOOST_ASSERT(IsValid());

			if (IsValid())
				return &it->car;
			else
				return NULL;
		}
		Car const & operator * (void) const
		{
			return this->operator *();
		}
		Car const * operator -> (void) const
		{
			return this->operator ->();
		}

		// -- constructeurs publiques --
		CarAcsor(CarAcsor const & op2)  // par copie
				: it(op2.it),
				  pValid(op2.pValid)
		{ }

		// la construction par defaut ne donne pas d'accesseur valide.
		CarAcsor(void)
		{ }

		CarAcsor operator = (CarAcsor const & op2)
		{
			it = op2.it;
			pValid = op2.pValid;
			return *this;
		}
		// comparaison
		bool operator == (CarAcsor const & op2) const
		{
			return it == op2.it;
		}

	private:
		CarIT it;  // iterateur de std::list.

		// une "classe d'instances" d'accesseur partage la même information de validité
		//   def : "classe d'instances", un groupe d'instances créés par copie ou assignation. (ici: des accesseurs référencants la même voiture)
		smart_ptr< bool > pValid;

		void InvalidateClass()
		{
			*pValid = false;
		}

		CarAcsor(CarIT a_it)	  // prive, mais CarManager peut en instancier
				: it(a_it),
				  pValid(new bool(true))   // seul ce constructeur prive peut valider un accesseur.
		{ }
	};

	// ajout d'une voiture dans la liste
	CarAcsor Push(char const* sprite);	// file name pour le sprite

	// suppression d'une voiture dans la liste (disparition silencieuse)
	void Pop(CarAcsor& acsor);  // retourne une copie de la voiture supprimée

	// même chose que "Pop" mais avec une explosion
	void Explode(CarAcsor& acsor, bool dontInvalidate = false);

	void DrawCars(void);

	void Update(double global_time, UPDATE_PARAM_en p = UP_TIME_ONLY);

	void SetExplosionSprite(char const* fname);
	void SetWreckageSprite(char const* fname);
	void SetCarPositionOverviewInfos(SpriteManager* pNoScrollSpriteMgr, int redid, int yellowid, Vector2i const& levelSize)
	{
		pos_view.redPixelId = redid;
		pos_view.yellowPixelId = yellowid;
		pos_view.pnoscroll_spm = pNoScrollSpriteMgr;
		pos_view.level_size = levelSize;
	}

	struct TooManyInstances {};

	CarsManager(void)
			: time(0.),
			  wreckId(-1),
			  explodeId(-1)
	{
		if (++instances != 1)
			throw TooManyInstances();
		prev_time = time;
		wrecks.resize(1000);  // maximum de 1000 épaves simultanées
		bombs.resize(50);	  // 50 explosions simultanées
	}

	~CarsManager()
	{
		--instances;
	}
};

#endif //CARSMANAGER_HPP_INC_19_03_2009___41607409_
