/**********************************************************************
 * File	: hostilesmanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 12/07/2007
 *********************************************************************/

#include <exception>
#include <algorithm>

#include "hostilesmanager.hpp"
#include "terraingenerator.hpp"
#include "collisionmanager.hpp"
#include "supramanager.hpp"
#include "iaclasses.hpp"

int HostilesManager::instances = 0;

namespace game_vars
{
	extern CarsManager::CarAcsor g_user;  // déclarée dans le module principal (solardiales.cpp)
	extern int g_minigun_id;
	extern int g_heavymg_id;
}

// ---------------------------------------------
// classe : cell
// ---------------------------------------------

void Cell::Pop(int idx)	 // operation en temps constant
{
	if (idx != ubound)
	{
		std::swap(nearlist[idx], nearlist[ubound]);

		// malheureusement, en bougeant le dernier element, on a change son indice dans l'information "Reverse"
		// il faut le mettre a jour:
		int i;
		for (i = 0; i < 4; ++i)	 // d'abord retrouver la cellule concernee
		{
			if (nearlist[idx].hostile->reverse[i].pcell == this)
			{
				nearlist[idx].hostile->reverse[i].index = idx;
				break;
			}
		}
	}
	--ubound;
}

int Cell::Push(Hostile* a_pHostile)
{
	++ubound;
	if (ubound >= membound)
	{
		membound = ubound * 2;	// redimensionnement exponentiel.
		nearlist.Resize(membound);
		// impossible d'utiliser une liste ni une freelist a cause des jeux d'indices
	}

	nearlist[ubound].hostile = a_pHostile;
	return ubound;	 // donner l'index a l'user
}


// ---------------------------------------------
// classe : grid
// ---------------------------------------------

void Grid::Initialize(int x, int y, int log2resolution)
{
	bitmask = 0xffffffff << log2resolution;  // pour s'aligner rapidement sur les cellules
	int wFittingOverSpan = 1;	// la grille doit deborder du terrain
	int hFittingOverSpan = 1;

	if ((x & bitmask) == x)	 // le terrain est deja un multiple de la resolution
		wFittingOverSpan = 0;
	if ((y & bitmask) == y)
		hFittingOverSpan = 0;

	//			x /  resolution
	width = (x >> log2resolution) + wFittingOverSpan;		// en nombre de cellules
	int height = (y >> log2resolution) + hFittingOverSpan;
	xmax = width << log2resolution;	// x max dans la grille en pixels
	ymax = height << log2resolution;
	nbCells = width * height;
	log2resol = log2resolution;
	grid = new Cell[++nbCells];  // + une pour tout l'espace exterieur.
}

void Grid::SignalMove(Hostile* phostile)
{
	int px[4], py[4];
	Cell* pcells[4];	 // cellules activees

	// verifier si on etait dans une seule cellule, en est-on sorti ?

	// calcul des deux coins significatifs : haut gauche, bas droite.

	// haut gauche:
	px[0] = phostile->pcar->rect.pos.x;
	py[0] = phostile->pcar->rect.pos.y;

	// bas droite:
	px[3] = px[0] + phostile->pcar->rect.w;
	py[3] = py[0] + phostile->pcar->rect.h;

	// comparaison des pointeurs de cellules activees :

	pcells[0] = GetCellPtr(px[0], py[0]);
	pcells[3] = GetCellPtr(px[3], py[3]);

	if (phostile->reverse[1].pcell == 0)  // si on etait dans une seule cellule
	{
		// et qu'on en sort pas :
		if (pcells[0] == phostile->reverse[0].pcell && pcells[3] == pcells[0])
			return;	// alors rien a faire
	}

	// on est sur une jonction.
	// on efface toutes les precedentes occurences de reference vers nous dans les cellules :

	phostile->EraseGridPresence(); // nettoyage

	// deuxieme etape, trouver les cellules avec lesquelles il rentre en contact
	// et les renseigner de sa presence.

	// deux derniers coins sur le rectangle : haut droite, bas gauche.

	// haut droite:
	px[1] = px[3];
	py[1] = py[0];

	// bas gauche:
	px[2] = px[0];
	py[2] = py[3];

	// trouver le reste des cellules activees :
	pcells[1] = GetCellPtr(px[1], py[1]);
	pcells[2] = GetCellPtr(px[2], py[2]);

	bool skip;
	int rev_i = 0;	 // indice independant pour acceder au talbeau des 4 reverse dans "hostile".
	for (int cell_i = 0; cell_i < 4; ++cell_i)  // pour chacune des 4 cellules qu'on a trouve associees aux 4 points
	{
		skip = false;
		for (int cl = 0; cl < cell_i; ++cl)	 // verifier si une precedente n'est pas la meme
		{
			if (pcells[cl] == pcells[cell_i])
			{
				skip = true;  // deja faite.
				break;
			}
		}
		if (!skip)
		{
			// renseignement de la reverse map : un pointeur vers la 'nouvelle' cellule
			phostile->reverse[rev_i].pcell = pcells[cell_i];
			phostile->reverse[rev_i].index = pcells[cell_i]->Push(phostile);	// ajout a la cellule
			// et l'indice ou on a ete ajoute dans cette cellule.
			++rev_i; // avancer independamment de cell_i car on peut tomber plusieurs fois sur la meme.
		}
	}
	for (int end = rev_i; end < 4; ++end)	// remplir la fin du tableau avec des NULL
	{
		phostile->reverse[end].pcell = 0;
	}
}

Grid::~Grid()
{
	if (grid)
		delete [] grid;
}


// ---------------------------------------------
// classe : HostilesManager
// ---------------------------------------------

bool HostilesManager::SpawnHostile(SPAWN_POSITION_en spawnPos, int isprite)
{
	// creation d'une voiture :
	CarsManager::CarAcsor newcar = ChiefManager::GetCarsManager()->Push(spriteFiles[isprite]);
	hostiles.push_front(newcar);
	// *hostiles.front().pcar et *newcar sont deux moyens d'acceder a la meme voiture
	const int bordw = ChiefManager::GetTerrainManager()->GetBorderWidth();
	const Vector2i terrSize = ChiefManager::GetTerrainManager()->GetSize();

	// faire spawner les ennemis a la toute droite du niveau
	if (spawnPos == SP_LEVEL_RIGHT)
	{
		// position aleatoire en vertical, mais 500 pixel a droite du bout du niveau
		newcar->SetPosition(terrSize.x + 500,
							bordw +
							rnd.Get(terrSize.y - bordw * 2));
	}
	else // MIDDLE
	{
		Rect rc(newcar->rect);

		// meme technique que dans le generateur de terrain
		int security = 0;
		bool intersect = true;
		const int max_repeats = terrSize.x / rc.w + terrSize.y / rc.h;	 // heuristique rapide
		while (intersect && security < max_repeats)	// jusqu'a que ca rentre
		{
			// choix aléatoire : position sur la carte
			rc.pos.x = bordw + 50 + rnd.Get(terrSize.x - rc.w - (bordw << 1));
			rc.pos.y = bordw + rnd.Get(terrSize.y - rc.h - (bordw << 1));
			// vérifier si on ne spawn pas sur un objet du decors:
			intersect = ChiefManager::GetCollisionsAndProbing()->IntersectRect(rc);
			++security;
		}
		if (security == max_repeats)	// pas marché
		{
			hostiles.pop_front();
			return false;
		}
		newcar->SetPosition(rc.pos.x,
							rc.pos.y);
	}
	// réglage de la cible :
	hostiles.front().target.x = 0;
	hostiles.front().target.y = 64 + rnd.Get(terrSize.y - 128);
	// angle aléatoire :
	newcar->angle = rnd.Get(359);

	newcar->GiveWeapon(game_vars::g_minigun_id);
	newcar->SetIAClass(iaClassID_simple, ChiefManager::GetIAManager());
	BOOST_ASSERT(newcar->GetIAClassID() == iaClassID_simple);

	if (isprite > 0)	// ennemi plus solide !
	{
		newcar->health = 200;
		newcar->SetIAClass(iaClassID_strong, ChiefManager::GetIAManager());
		BOOST_ASSERT(newcar->GetIAClassID() == iaClassID_strong);

		newcar->GiveWeapon(game_vars::g_heavymg_id);   // et plus dangereux !
	}

	return true;
}

void HostilesManager::SetIAClassIDs(size_t classSimple, size_t classStrong)
{
	iaClassID_simple = classSimple;
	iaClassID_strong = classStrong;
}

void HostilesManager::RegisterPlayer(const CarsManager::CarAcsor& user)
{
	hostiles.push_front(user);
	hostiles.front().isplayer = true;
	++playersRegisteredNum;
}

void HostilesManager::SetSpawnRule(int spawnRules_flag, int flagParameter)
{
	options = spawnRules_flag;
	if (options & SPR_FREQUENCY)
	{
		frequency = flagParameter;
	}
}

void HostilesManager::InitGrid(void)
{
	if (gridNotInit)
	{
		grid.Initialize(ChiefManager::GetTerrainManager()->GetSize().x, ChiefManager::GetTerrainManager()->GetSize().y);
		gridNotInit = false;
	}	
}

void HostilesManager::Update(float delta_time, Vector2f a_screenCenter, bool spawnAdvancedUnits)
{
	now += delta_time;

	if (gridNotInit)
	{
		InitGrid();
	}

	// règles de spawns aléatoires dans le temps :
	timeAccumulator += delta_time;
	if (!(options & SPR_DISABLED))
	{
		// si on dépasse le temps prévu pour le prochain spawn, alors on spawn :
		if (timeAccumulator >= nextSpawn && spriteFiles.size() > 0)
		{
			// determiner le temps du prochain spawn :
			if (frequency == 0)
				nextSpawn = 1e37f;
			else
				nextSpawn = rnd.Get(12000 / frequency) * 0.01f;	 // frequency est un nombre d'ennemis par minute moyen.
			timeAccumulator = 0.f;	// reset

			bool mid = options & SPR_MIDDLE;
			bool rig = options & SPR_RIGHT;
			if (mid && rig) // il faut decider lequel
			{
				if (rnd.Get(1) == 1)
					rig = false;
				else
					mid = false;
			}
			if (rig)	 // a droite
			{
				SpawnHostile(SP_LEVEL_RIGHT, spawnAdvancedUnits ? 1 : 0);
			}
			if (mid)	 // au milieu
			{
				SpawnHostile(SP_LEVEL_MIDDLE, spawnAdvancedUnits ? 1 : 0);
			}
		}
	}

	screenCenter = a_screenCenter;

	ChiefManager::GetProximityManager()->Reset();   // vider la liste des proximités car elle va etre reconstruite

	// calculer les mouvements de chaque hostile
	list< Hostile >::iterator it;
	for (it = hostiles.begin(); it != hostiles.end(); ++it)
	{
		IntelligenceAndReaction(*it);

		it->pcar->Update(delta_time);
		grid.SignalMove(&(*it));

		// explosion : si vie négative ou nulle
		if (it->pcar->health <= 0)
		{
			ChiefManager::GetWeaponManager()->CeaseFire(&(*it->pcar));
			ChiefManager::GetCarsManager()->Explode(it->pcar, it->isplayer);  // arg2: on ne sort pas le joueur du CarManager quand il meurt
			if (it->isplayer)
				--playersRegisteredNum;
			// on efface l'hostile de la liste :
			it->EraseGridPresence();
			it = hostiles.erase(it);
			--it;
		}
		// règle de disparition : arriver près du bord
		else if ((it->pcar->rect.pos.x <= ChiefManager::GetTerrainManager()->GetBorderWidth() + 50)
				 && !it->isplayer)
		{
			ChiefManager::GetWeaponManager()->CeaseFire(&(*it->pcar));
			ChiefManager::GetCarsManager()->Pop(it->pcar);	// on efface la voiture du manager.
			it->EraseGridPresence();	// et l'hostile.
			it = hostiles.erase(it);
			--it;
		}
		else if (!it->isplayer)  // ni explosion, ni pres du bord, et n'est pas le joueur
		{
			// signaler l'ennemi au gestionnaire de proximité
			float distanceToPlayerSq = DistTwoPtsSq(it->pcar->rect.GetCenter(), game_vars::g_user->rect.GetCenter());
			ChiefManager::GetProximityManager()->ConditionalAdd(it->pcar, distanceToPlayerSq, PM_SQUARED_DIST);
		}
	}
}

void HostilesManager::IntelligenceAndReaction(Hostile& hostile)
{
	BOOST_ASSERT(hostile.pcar.IsValid());

	if (hostile.isplayer)
		return;

	// quelques variables utiles
	Vector2i const mypos = hostile.pcar->rect.GetCenter();
	Vector2i const usrpos = game_vars::g_user->rect.GetCenter();
	bool const usralive = game_vars::g_user->health > 0;
	Vector2f& mydir = hostile.pcar->direction;
	Hostile::IADataCache& ia_data = hostile.iaDataCache;

	if (hostile.CheckRefreshIA())
	{
		Vector2i touser = mypos - usrpos;
		// l'ennemi est à moins de 220 pixels du joueur (en norme 1):
		bool isnear = touser.LenghtOne() < 220;
		// on est face a l'ennemi:
		bool isvisible = Vector2f(touser) * mydir < 0;
		if (isnear && isvisible && !ia_data.pursuit.active && usralive && !hostile.CheckOverride(now))
		{
			ia_data.pursuit.active = true;   // on active la poursuite
			ia_data.pursuit.startTime = now;
			// et on ouvre le feu !
			int weaponid = hostile.pcar->WeaponPresent(game_vars::g_heavymg_id) ? game_vars::g_heavymg_id : game_vars::g_minigun_id;
			ChiefManager::GetWeaponManager()->OpenFire(&(*hostile.pcar), weaponid);
		}
		else if (( !isnear || !usralive ) // trop loin ou bien le joueur est mort
				 && ia_data.pursuit.active)
		{
			ia_data.pursuit.active = false;  // désactiver la poursuite
			ChiefManager::GetWeaponManager()->CeaseFire(&(*hostile.pcar));
		}
		else if (ia_data.pursuit.active)  // juste en mode poursuite
		{
			bool too_close = touser.LenghtOne() < 100;
			if (too_close)
			{
				Car::DIRECTION_en dir = (time(0) % 2) ? Car::DIR_RIGHT : Car::DIR_LEFT;
				hostile.pcar->SetDirection(dir);
				hostile.ActiveOverride(now, 3.f);	// demi tour
				ia_data.pursuit.active = false;  // désactiver la poursuite
				ChiefManager::GetWeaponManager()->CeaseFire(&(*hostile.pcar));
			}
		}
	}
	if (!hostile.CheckOverride(now))	 // si on est en override, on suit le vieil ordre tant qu'il n'est pas fini.
	{
		Vector2i target = hostile.target;
		if (ia_data.pursuit.active)
		{
			target = usrpos;
		}
		// intelligence de base : se diriger vers la cible
		Vector2i toTarget(target - mypos);
		Vector2i ortho(toTarget.GetOrtho());

		// on projette la direction actuelle de la voiture sur le vecteur qui va vers la cible
		// puis on va a gauche si positif, et a droite si negatif.
		float proj = mydir * Normalize(Vector2f(ortho));
		float reverse = mydir * toTarget;         // si positif, bonne direction, sinon, reverse.
		if (proj > 0.1f)
		{
			hostile.pcar->SetDirection(Car::DIR_LEFT);
		}
		else if (proj < -0.1f)
		{
			hostile.pcar->SetDirection(Car::DIR_RIGHT);
		}
		else if (reverse < 0) // la on va pas dans le bon sens.
		{
			hostile.pcar->SetDirection(Car::DIR_LEFT);
		}
		else	// parfait
		{
			hostile.pcar->SetDirection(Car::DIR_STRAIGHT);
		}

		if (hostile.RefreshIA())  // juste une alternance, pour alléger le programme on check une image sur 4.
		{
			// évitement d'obstacle :

			if (hostile.StuckWatchDog(now))	// système anti coincement
			{	// on est coincé
				Car::DIRECTION_en chosenpath = ((time(0) >> 5) % 2) ? Car::DIR_RIGHT : Car::DIR_LEFT;   // inverse toutes les 32 secondes
				hostile.pcar->SetDirection(chosenpath);
				hostile.ActiveOverride(now, 2.5f);	// demi tour
				return;	// sert a rien d'envoyer les rayons
			}

			// lancer trois rayons devant, prendre le plus petit en dessous d'un certain seuil
			// tourner du cote opose
			Vector2f start1, start2, start3;
			Vector2f v;
			start1 = mypos;
			v = mydir * 5;
			start1 -= v;
			float dist1 = 1e37f;

			start3 = start2 = start1;
			start1 += mydir.GetOrtho() * 12;  // a peu près les "phares" de la voiture
			start2 -= mydir.GetOrtho() * 12;
			float dist2 = 1e37f;
			float dist3 = 1e37f;
			ChiefManager::GetCollisionsAndProbing()->IntersectRay(&dist1, start1, mydir, true, true);
			ChiefManager::GetCollisionsAndProbing()->IntersectRay(&dist2, start2, mydir, true, true);
			ChiefManager::GetCollisionsAndProbing()->IntersectRay(&dist3, start3, mydir, true, true);

			if (dist3 < 30) // probablement coincé
			{
				hostile.pcar->SetDirection(Car::DIR_LEFT);
				hostile.ActiveOverride(now, 2.f);  // demi tour
			}
			else
			{
				if (dist1 < dist2)
				{
					if (dist1 < 80)
					{
						hostile.pcar->SetDirection(Car::DIR_LEFT);
						hostile.ActiveOverride(now, 0.5f);
					}
				}
				else
				{
					if (dist2 < 80)
					{
						hostile.pcar->SetDirection(Car::DIR_RIGHT);
						hostile.ActiveOverride(now, 0.5f);
					}
				}
			}
		} // end if refreshIA
	} // end if overriden
}

// detection de collision entre un point et un ennemi :
//	  utile pour les tirs de roquettes
bool HostilesManager::HostileIntersectPoint(Car** a_ppcar, const Vector2i& point)
{
	// récuperation de la cellule activée par le point :
	Cell* pc = grid.GetCellPtr(point.x, point.y);

	// et boucler pour chaque ennemi contenu dans cette cellule :
	for (int enn = 0; enn <= pc->ubound; ++enn)
	{
		// création du orect :
		ORect eor = pc->nearlist[enn].hostile->pcar->GetORect();
		if (eor.IsPointInside(point))	 // si collision
		{
			*a_ppcar = &(*pc->nearlist[enn].hostile->pcar);
			return true;
		}
	}
	return false;
}

// détection de collision entre un rayon et un ennemi :
//	 utile pour tirer dessus par exemple.
bool HostilesManager::HostileIntersectRay(float* t, Car** a_ppcar, const Vector2f& pos, const Vector2f& dir)
{
	// detection des cellules activees par le rayon dans un proche voisinage
	Cell* activates[4];	// activons en 4 maxi.

	int x = (int)pos.x;
	int y = (int)pos.y;

	int cellw = grid.GetCellWidth();
	int mask = grid.GetBitmask();

	// cas particulier a traiter separament du fait des coefficients directeurs de droite
	// infinis (en x ou y)
	int xfactor = 0;
	int yfactor = 0;
	bool special = false;
	if (myutils::abs(dir.y) <= 1e-6f)  // si presque 0 (rayon horizontal)
	{
		if (dir.x > 0)	 // vers la droite
			xfactor = cellw;
		else
			xfactor = -cellw;	 // vers la gauche.
		special = true;
	}
	else if (myutils::abs(dir.x) <= 1e-6f)	 // rayon presque vertical
	{
		if (dir.y > 0)	 // vers le bas
			yfactor = cellw;
		else
			yfactor = -cellw;	 // vers le haut
		special = true;
	}
	if (special)  // on check 3 cases
	{
		for (int i = 0; i < 3; ++i)
		{
			activates[i] = grid.GetCellPtr(x + i * xfactor, y + i * yfactor);
		}
		activates[3] = 0;	 // la derniere inactivee.
	}
	else	// cas normal (rayon diagonal)
	{
		float coef = dir.y / dir.x;
		float invcoef = dir.x / dir.y;  // inverse du coefficient directeur
		int limit = 4;
		int h;
		int w;
		float dx, dy;

		for (int cel = 0; cel < limit; ++cel)
		{
			// determiner le x du croisement entre rayon et la prochaine ligne horizontale de la grille

			// coordonnee de la ligne horizontale a tester :
			h = y & mask;
			w = x & mask;
			if (dir.y > 0)	  // si le rayon est descendant, on teste la ligne du bas
				h += cellw;
			// deplacement en y pour atteindre cette ligne :
			dy = h - y;
			// deplacement en x engendre par un tel deplacement en y en suivant le coeficient directeur :
			dx = dy * invcoef;

			activates[cel] = grid.GetCellPtr(x, y);

			if (x + dx > w + cellw)	 // debordement lateral positif
			{
				dx = w + cellw - x;	// on prend le dx qui va a la prochaine cellule
				x = w + cellw;
				y = y + (int)(coef * dx);	// pour trouver le dy
			}
			else if (x + dx < w)	 // lateral negatif
			{
				dx = x - w;
				x = w - 1;	// pour activer la cellule precedente il faut au moins rentrer dedans d'un pixel
				y = y - (int)(coef * dx);
			}
			else	// vertical
			{
				x = (int)(x + dx);
				y = h;
				if (dir.y < 0)
					--y;	// pareil, il faut rentrer dans la case du dessus.
			}
		}
	}
	// compteur du nombre d'appels a cette fonction. (partagé entre instances mais ca n'a aucune importance)
	static unsigned int calls = 0;
	++calls;
	float myt, mint = 1e37f;
	bool hit = false;
	Car* pcar = 0;
	for (int cel = 0; cel < 4 && activates[cel] != 0; ++cel)	 // pour chaque cellules activees
	{
		Cell& c = *activates[cel];
		for (int ls = 0; ls <= c.ubound; ++ls)	 // pour chaque ennemi contenu dans cette cellule
		{
			if (c.nearlist[ls].checked != calls)  // si pas deja traite
			{
				c.nearlist[ls].checked = calls;
				// creation du orect :
				ORect eor = c.nearlist[ls].hostile->pcar->GetORect();
				if (eor.IntersectsRay(&myt, pos, dir)) // si collision
				{
					if (myt < mint && myt >= 0)  // est-ce qu'elle est plus proche que la plus proche precedemment detectee
					{
						mint = myt;	 // si oui on la prend comme nouvelle plus proche
						pcar = &(*c.nearlist[ls].hostile->pcar);
					}
					hit = true;	 // oui on a touche qqch
				}
			}
		}
	}
	if (hit)
	{
		*t = mint;	// on renvoie la position a l'utilisateur
		*a_ppcar = pcar;	// et la voiture touchee
	}
	return hit;
}

// methode pour controller la collision entre un rectangle et les ennemis qui se trouvent dans ce rectangle
// utile pour detruire les ennemis qui se trouveraient par megarde sous une tourelle lors de sa construction
//	 note : prend en argument une reference vers une liste de Car*

//	 ATTENTION : limitation de la fonction a des rectangles plus petits qu'une cellule de partitionnement

bool HostilesManager::HostileIntersectRect(list< Car* >& a_lrcars, const Rect& rc)
{
	// securite : (on est jamais trop prudent)
	//	 de plus cette fonction peut se permettre d'etre lente (casser le pipeline) car appellee rarement.
	if (rc.w > grid.GetCellWidth() || rc.h > grid.GetCellWidth())
		throw std::invalid_argument("arguments outside of limits for HostileIntersectRect function");

	Cell* pcells[4];	 // cellules activees

	// comparaison des pointeurs de cellules activees :

	pcells[0] = grid.GetCellPtr(rc.pos.x,			rc.pos.y);
	pcells[3] = grid.GetCellPtr(rc.pos.x + rc.w, rc.pos.y + rc.h);

	int bound = 3;
	if (pcells[0] == pcells[3])  // si rc est dans une seule cellule
	{
		bound = 0;
	}
	else
	{
		// il faut determiner les autres cellules
		pcells[1] = grid.GetCellPtr(rc.pos.x + rc.w, rc.pos.y);
		pcells[2] = grid.GetCellPtr(rc.pos.x,			rc.pos.y + rc.h);
		if (pcells[0] == pcells[1])  // que deux cellules en hauteur
		{
			pcells[1] = pcells[2];
			bound = 1;	// on limite la recherche a 2 cellules
		}
		else if (pcells[0] == pcells[2])	 // que deux cellules en largeur
		{
			bound = 1;	// pareil
		}
	}
	// 0 -- 1
	// |	  |
	// 2 -- 3

	bool anythingEncountered = false;
	// boucler pour les cellules selectionees:
	for (int cl = 0; cl <= bound; ++cl)
	{
		// et boucler pour chaque ennemi contenu dans la cellule cl :
		for (int enn = 0; enn <= pcells[cl]->ubound; ++enn)
		{
			// création du orect :
			ORect eor = pcells[cl]->nearlist[enn].hostile->pcar->GetORect();
			if (eor.IntersectsORc(rc))	 // si collision
			{
				// ajouter le pointeur de la voiture en collision a la liste (argument user) :
				a_lrcars.push_back( &(*pcells[cl]->nearlist[enn].hostile->pcar) );
				anythingEncountered = true;
			}
		}
	}
	return anythingEncountered;
}

// cette fonction va, lors de la construction d'une tourelle, construire une liste de
// cellules activees par la presence de la tourelle qui aidera a l'acceleration
// de la recherche des ennemis a portee de tir.
HostilesManager::FireRange HostilesManager::BuildFireRangeCellList(const Vector2i& center, int radius)
{
	// en faisant une double boucle on parcours toutes les cellules dans un carre
	// englobant virtuel autours du centre.

	FireRange result;

	int cw = grid.GetCellWidth();
	int nums = (radius * 2) / cw;

	// point haut gauche :
	Vector2i hlcorner = center;
	hlcorner.x -= cw * (nums / 2);
	hlcorner.y -= cw * (nums / 2);

	// si nums est un multiple de 2 alors le carre englobant sera forcement de-centre.
	// il faut donc l'ajuster plus finement en regardant	si le depart de tir
	// est au dela de la moitie de la case, alors on shift le sampling des cases:
	if (!(nums & 1))
	{
		if ( (center.x & ~grid.GetBitmask()) > (cw >> 1) )
			hlcorner.x += cw;
		if ( (center.y & ~grid.GetBitmask()) > (cw >> 1) )
			hlcorner.y += cw;
	}

	Vector2i activation;
	for (int x = 0; x < nums; ++x)
	{
		for (int y = 0; y < nums; ++y)
		{
			// calculer le point d'activation de cellule :
			activation.x = hlcorner.x + x * cw;
			activation.y = hlcorner.y + y * cw;
			// stocker la cellule activee:
			result.pcellList->Add( grid.GetCellPtr(activation.x, activation.y) );
		}
	}

	result.firecenter = center;
	result.maxradius = radius;
	return result;
}

// chercher un ennemi proche et en ligne de mire (sans obstacles)
// dans une zone de tir (range).
Car* HostilesManager::GiveMeAnEnnemiInRange(const FireRange& range, ia::Class const & callerClass)
{
	// boucler sur les ennemis compris dans les cellules du champ de tir.

	float distance = 1e37f;
	Cell* pcl;
	Car* pcar = 0;
	Car* candidate = 0;
	Vector2i epos;

	// pour toutes les cellules dans la liste de proximite :
	for (int cl = 0; cl < range.pcellList->GetElementsNumber(); ++cl)
	{
		pcl = range.pcellList->GetElement(cl);	 // raccourcir l'ecriture
		// pour chaque ennemi contenu dans la cellule pcl
		for (int enn = 0; enn <= pcl->ubound; ++enn)
		{
			pcar = &(*pcl->nearlist[enn].hostile->pcar);
			epos = pcar->rect.GetCenter();  // position centrale de l'ennemi
			float sqd = DistTwoPtsSq(range.firecenter, epos);
			// si on est ennemi avec l'hostile visé, et qu'il est pas trop loin:
			if (callerClass.RelationWith(*pcar, ChiefManager::GetIAManager()) == ia::ENNEMY &&
				sqd < distance && sqd < range.maxradius * range.maxradius)
			{
				// verifier que le champ de tir est ouvert (pas d'obstacles)
				bool accept = true;
				float t;
				Vector2f dir = Normalize(epos - range.firecenter);
				if (ChiefManager::GetCollisionsAndProbing()->IntersectRay(&t, range.firecenter, dir, false, false))
				{
					t = t * t;	// on compare des distances au carre
					if (t < sqd)  // collision
						accept = false;
				}
				// si c'est bon on garde cet ennemi en memoire
				if (accept)
				{
					distance = sqd;
					candidate = pcar;
				}
			}
		}
	}
	// retourner l'ennemi choisi
	return candidate;
}

// ------------------------------------
// classe : Hostile
// ------------------------------------

void Hostile::ActiveOverride(double globalTime, float duration)
{
	if (!overrideOn)
	{
		overrideOn = true;
		stamp = globalTime;
		overrideDuration = duration;
	}
}

bool Hostile::CheckOverride(double globalTime)
{
	if (overrideOn)
	{
		if (stamp + overrideDuration < globalTime)
		{
			overrideOn = false;
		}
	}
	return overrideOn;
}

void Hostile::EraseGridPresence(void)
{
	// referencer les cellules a effacer
	// en parcourant la reverse map contenue dans l'hostile lui meme !

	for (int i = 0; i < 4; ++i)  // pour chaque cellule
	{
		Cell* pcell = reverse[i].pcell;
		if (pcell != 0)
		{
			pcell->Pop(reverse[i].index);	 // pop
			reverse[i].pcell = 0;
		}
	}
}
