/**********************************************************************
 * File	: hostilesmanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 12/07/2007
 *********************************************************************/

#ifndef HOSTILESMANAGER_HPP_INC_19_03_2009__100900170_
#define HOSTILESMANAGER_HPP_INC_19_03_2009__100900170_ 1

#include "carsmanager.hpp"
#include "proximity_manager.hpp"

#include "randclass.hpp"
#include "SecureArray.hpp"
#include "smart_ptr.hpp"

#include "geometry.hpp"
#include "csttimegetsizelist.hpp"

#include <vector>
using std::vector;
#include <list>
using std::list;

namespace ia
{
	class Class;
}

enum SPAWN_POSITION_en
{
	SP_LEVEL_RIGHT,
	SP_LEVEL_MIDDLE
};

enum SPAWN_RULES_en
{
	// option flags
	SPR_DISABLED  = 1,
	SPR_RIGHT	  = 2,
	SPR_MIDDLE	  = 4,
	SPR_BOTH	  = 6,
	SPR_FREQUENCY = 8	 // hostiles par minute
};

// structure pour accelerer la recherche des ennemis.
//	 -> limiter les tests de collisions

class Hostile;

// cellule de grille.
// une cellule contient la liste des ennemis s'y trouvant.
class Cell
{
	int membound;

public:

	struct HostileRef
	{
		Hostile* hostile;
		unsigned int checked;  // ici aussi on met un compteur d'accès

		HostileRef(void)
				: checked(0) { }
	};

	//	  -- publique pour pouvoir iterer dessus --
	// (oui j'ai pas le temps d'investiguer sur comment fournir des iterateurs)
	SecureArray< HostileRef > nearlist;
	int ubound;

	void Pop(int idx);
	int Push(Hostile* a_pHostile);

	Cell(void)
			: membound(10),
			  ubound(-1)
	{
		nearlist.Resize(10);	  // deja 600ko a l'initialization si la carte fait 200,200...
	}
};

// subdivise l'espace en sous carrés de taille constante
// maintient des listes d'ennemis
class Grid
{
	Cell* grid;
	int log2resol;
	int bitmask;
	int xmax;  // en pixels
	int ymax;
	int nbCells;	 // nb total de cellule, y compris celle sans position
	int width;	// nb de cellules en largeur

public:

	Grid(void) : grid(0) { }

	void Initialize(int x, int y, int log2resolution = 8);	// zone a couvrir

	void SignalMove(Hostile* phostile);

	int GetLog2resol(void)
	{
		return log2resol;
	}

	int GetCellWidth(void)
	{
		return 1 << log2resol;
	}

	int GetBitmask(void)
	{
		return bitmask;
	}

	Cell* GetCellPtr(int x, int y)
	{
		if (x >= xmax || y >= ymax || x < 0 || y < 0)
			return grid + nbCells - 1;	 // cellule speciale "exterieur"

		x >>= log2resol;	 // division par la resolution pour trouver les indices de la cellule
		y >>= log2resol;

		return grid + (x + y * width);
	}

	~Grid();
};

class Hostile
{
	// variables privées

	float stamp;
	bool overrideOn;
	float overrideDuration;
	int iaCount;

	Vector2i wtch_oldpos;
	float wtch_time;

public:

	bool isplayer;
	CarsManager::CarAcsor pcar;
	Vector2i target;

	// stocke où est-ce que j'ai été placé dans la grille et dans chaque cellule :
	struct ReverseGrid
	{
		Cell* pcell;
		int index;

		ReverseGrid(void)
				: pcell(0) { }
	} reverse[4];     // étant donné que les cellules sont plus grandes que les hostiles
	                  // on ne peut pas être dans plus de 4 cellules à la fois.

	struct IADataCache
	{
		struct Pursuit
		{
			bool active;
			double startTime;

			Pursuit() : active(false) {}
		} pursuit;

// 		IADataCache()
// 		{}
	} iaDataCache;

	void ActiveOverride(double gtime, float duration);
	bool CheckOverride(double gtime);

	bool CheckRefreshIA(void) const
	{
		return iaCount - 1 == 0;
	}
	bool RefreshIA(void)
	{
		if (--iaCount == 0)
		{
			iaCount = 3;
			return true;
		}
		return false;
	}

	// chien de garde anti coincage
	bool StuckWatchDog(double globalTime)
	{
		bool stuck = false;
		if (globalTime > wtch_time + 3)	 // toutes les 3 secondes on verifie qu'on est pas coince
		{
			if ((pcar->rect.pos - wtch_oldpos).LenghtSq() <= 9)  // si on a bouge de moins-ou-de 3 pixels
			{
				stuck = true;	// on est coince
			}
			wtch_time = globalTime;
			wtch_oldpos = pcar->rect.pos;
		}
		return stuck;
	}

	void EraseGridPresence(void);

	Hostile(CarsManager::CarAcsor const& a_pcar)
			: overrideOn(false),
			  iaCount(1),
			  wtch_oldpos(-50, -50),
			  wtch_time(0),
			  isplayer(false),
			  pcar(a_pcar)
	{}
};

class HostilesManager
{
	tools::CstTimeGetSizeList< Hostile > hostiles;
	Grid grid;

	tools::Random rnd;
	int options;
	int frequency;
	double timeAccumulator;
	float nextSpawn;
	bool gridNotInit;
	Vector2f screenCenter;
	double now;
	size_t iaClassID_simple;
	size_t iaClassID_strong;
	size_t playersRegisteredNum;

	static int instances;
	struct TooManyInstances { };

public:

	// un fire range est une zone ou une tourelle va pouvoir chercher des ennemis a cibler.
	class FireRange
	{
		friend class HostilesManager;

		// concretement, on veux une liste de cellules ou boucler pour chercher ses ennemis
		// sauf que pour aller vite il vaut mieux stocker un tableau, vu que la collection
		// n'est consruite qu'une fois a l'intanciation de la tourelle;
		// et qu'en revanche, elle sera parcourue souvent.
		// et finalement on veut que la fonction retourne le resultat (plus joli que passer une reference)
		// donc un smart pointer permet d'allouer dans la fonction et de ne pas se preoccuper de la destruction

		typedef SecureArray< Cell* > CellPtrArray;
		smart_ptr< CellPtrArray > pcellList;

		Vector2i firecenter;
		float maxradius;

		// le tout est prive car l'utilisateur n'a que faire de ce que contient un FireRange.
		// par contre HostilesManager DOIT pouvoir lire le contenu pour travailler dessus -> donc amie.

	public:

		FireRange(void)
				: pcellList(new CellPtrArray)
		{ }
	};

	vector< const char* > spriteFiles;

	// retourne faux si il n'a pas ete possible d'inserer un ennemi dans le niveau
	bool SpawnHostile(SPAWN_POSITION_en spawnPos, int indexSprite);

	void SetIAClassIDs(size_t classSimple, size_t classStrong);

	void RegisterPlayer(const CarsManager::CarAcsor& user);

	void SetSpawnRule(int spawnRules_flag, int flagParameter = 0);

	void InitGrid(void);

	void Update(float global_time, Vector2f screenCenter, bool SpawnAdvancedUnits);

	void IntelligenceAndReaction(Hostile& hostile);

	bool HostileIntersectPoint(Car** a_ppcar, const Vector2i& point);

	bool HostileIntersectRay(float* t, Car** a_ppcar, const Vector2f& pos, const Vector2f& dir);

	bool HostileIntersectRect(list< Car* >& a_lrcars, const Rect& rc);

	FireRange BuildFireRangeCellList(const Vector2i& center, int radius);

	Car* GiveMeAnEnnemiInRange(const FireRange& range, ia::Class const &);

	int GetEnnemiesNum(void)
	{
		return hostiles.size() - playersRegisteredNum;
	}

	HostilesManager(void)
			:  timeAccumulator(0.),
			   nextSpawn(1.f),
			   gridNotInit(true),
			   now(0.),
			   iaClassID_simple(777777),   // parce que c'est facile a voir dans le debuggeur
			   iaClassID_strong(777777),
			   playersRegisteredNum(0)
	{
		if (++instances != 1)
			throw TooManyInstances();
		SetSpawnRule(SPR_BOTH | SPR_FREQUENCY, 6);
	}
	~HostilesManager()
	{
		--instances;
	}
};

#endif //HOSTILESMANAGER_HPP_INC_19_03_2009__100900170_
