/**********************************************************************
 * File	: hud.cpp
 * Author: Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 15/07/2007
 *********************************************************************/

#include "hud.hpp"

#include <boost/assert.hpp>

#include <stdio.h>
#include <stdexcept>

#ifdef WIN32
#define snprintf _snprintf
#endif

//-------------------------------------------------------
// Menu
//-------------------------------------------------------

int Menu::PushLeft(const char* sprite)
{
	MenuElement elem;
	elem.sprId = spm.LoadSprite(sprite, SP_STATIC);	  // charger le sprite representant l'element.
	elem.height = spm.GetH(elem.sprId);
	int nextFreeY;
	if (l_elements.size())
		nextFreeY = l_elements.back().posy + l_elements.back().height + spacing;
	else
		nextFreeY = 0;
	elem.posy = nextFreeY;
	l_elements.push_back(elem);
	spm.Move(elem.sprId, lx, ly + elem.posy);
	return -l_elements.size() + 1;
}

int Menu::PushRight(const char* sprite)
{
	MenuElement elem;
	elem.sprId = spm.LoadSprite(sprite, SP_STATIC);	  // sprite id
	elem.height = spm.GetH(elem.sprId);			// hauteur
	int nextFreeY;
	if (r_elements.size())
		nextFreeY = r_elements.back().posy + r_elements.back().height + spacing;
	else
		nextFreeY = 0;
	elem.posy = nextFreeY;
	r_elements.push_back(elem);	 // ajout dans le menu
	spm.Move(elem.sprId, rx, ly + elem.posy);		// placement du sprite
	return r_elements.size();
}

int Menu::GetLeftSize(void)
{
	return l_elements.size();
}

int Menu::GetRightSize(void)
{
	return r_elements.size();
}

Vector2i Menu::GetElemPos(int id)
{
	if (id > 0)
	{
		return Vector2i(rx,
						ry + r_elements[id - 1].posy + r_elements[id - 1].height / 2);
	}
	else
	{
		return Vector2i(lx,
						ly + l_elements[-id].posy + l_elements[-id].height / 2);
	}
}

Vector2i Menu::GetElemSize(int id)
{
	vector< MenuElement >& elemCtnr = id > 0 ? r_elements : l_elements;
	id = id > 0 ? id - 1 : -id;
	return Vector2i(spm.GetW(elemCtnr[id].sprId),
					elemCtnr[id].height);
}

void Menu::SignalCursorMove(MENU_CURSOR_MOVE_en direction)
{
	static std::map< MENU_CURSOR_MOVE_en, int > mvts;
	if (mvts.empty())
	{
		mvts[M_CUR_LEFT]  = -1;
		mvts[M_CUR_STAY]  = 0;
		mvts[M_CUR_RIGHT] = +1;
	}
	SetCursorPosition(cursorPos + mvts[direction]);
}


void Menu::PlaceArrow(void)
{
	int cw = spm.GetW(arrowId) + 5;	// largeur du curseur + 5 px de marge
	int ch = spm.GetH(arrowId);
	Vector2i pos = GetElemPos(cursorPos);
	spm.Move(arrowId,
			 pos.x - cw,
			 pos.y - ch / 2);
}

void Menu::SetRightStartPos(int x, int y)
{
	rx = x;
	ry = y;
}

void Menu::SetLeftStartPos(int x, int y)
{
	lx = x;
	ly = y;
}

void Menu::SetVerticalSpacing(int s)
{
	spacing = s;
}

void Menu::SetArrow(const char* fn_arrow)
{
	arrowId = spm.LoadSprite(fn_arrow);
	spm.SetSpeed(arrowId, 6);
	SignalCursorMove(M_CUR_STAY);
}

int Menu::SetHeadLeft(int x, int y, const char* sprite)
{
	headleft = spm.LoadSprite(sprite, SP_STATIC);
	spm.Move(headleft, x, y);
	return headleft;
}

int Menu::SetHeadRight(int x, int y, const char* sprite)
{
	headright = spm.LoadSprite(sprite, SP_STATIC);
	spm.Move(headright, x, y);
	return headright;
}

int Menu::GetCursorPosition(void)
{
	return cursorPos;
}

void Menu::SetCursorPosition(int pos)
{
	int lowval = -l_elements.size() + 1;
	int highval = r_elements.size();
	if (pos < lowval)
		cursorPos = highval;
	else if (pos > highval)
		cursorPos = lowval;
	else
		cursorPos = pos;
	PlaceArrow();
}

void Menu::Draw(EDrawType dt /*= WITH_FLIP*/)
{
	for (int i = 0; i < spm.GetNumSprites(); ++i)
	{
		if (i == headleft && cursorPos > 0) // a droite -> on dessine pas la tete de gauche
		{
		}
		if (i == headright && cursorPos <= 0)	// a gauche -> reciproquement.
		{
		}
		else
		{
			spm.Draw(i);
		}
	}
	spm.Update(dt == WITH_FLIP ? UP_SCREEN_AND_TIME : UP_TIME_ONLY);
}

//-------------------------------------------------------
// Numbers
//-------------------------------------------------------

void Numbers::PrintNum(int x, int y, unsigned int value)
{
	// conversion valeur -> chaine
	char str[16];
	snprintf(str, 15, "%u", value);
	// valeur du caractere 0:
	char zero = '0';	 // pour le reste on suppose que ca se suit dans la table de code.
	int i = 0;
	int toprint;
	int xpos = x;
	while (str[i] != '\0')
	{
		toprint = str[i] - zero;
		spm.Draw(0, xpos, y, (float)toprint);
		++i;
		xpos += w;
	}
}

//-------------------------------------------------------
// Hud
//-------------------------------------------------------

int Hud::PushNewHudElement(int x, int y, const char* fname, int a_numberFontID, bool animated)
{
	Element elem;
	elem.x = x;
	elem.y = y;
	elem.numberFontID = a_numberFontID;
	if (a_numberFontID < 0)  // si ce n'est pas un nombre
	{
		SPRITE_PARAM_en arg = animated ? SP_ANIMATED : SP_STATIC;
		elem.spIds.push_back(sprMngr.LoadSprite(fname, arg));
	}
	elements.push_back(elem);
	return elements.size() - 1;
}

int Hud::PushAlternateHudElement(int baseElement, const char* fname)
{
	if (baseElement >= 0 && baseElement < (signed)elements.size())
	{
		elements[baseElement].spIds.push_back(sprMngr.LoadSprite(fname, SP_STATIC));
		return elements[baseElement].spIds.size() - 1;
	}
	else
		return -1;
}

void Hud::SetElementVisible(int elemId, bool visible)
{
	elements[elemId].visible = visible;
}

void Hud::UpdateNumberValue(int elemId, unsigned int value)
{
	elements[elemId].value = value;
}

void Hud::UpdateAlternateElement(int baseElement, int alternateElement)
{
	BOOST_ASSERT((signed)elements[baseElement].spIds.size() > alternateElement && alternateElement >= 0);

	elements[baseElement].current = alternateElement;
}

int Hud::PushNewNumbersFontSprite(const char* file)
{
	numbers.push_back(smart_ptr<Numbers>(new Numbers()));
	numbers.back()->LoadNumbers(file);
	return numbers.size() - 1;
}

int Hud::PushProgressBar(int x, int y, const char* bar, const char* criticalBar, EPBarType etype /*=PB_STANDARD*/)
{
	// 350 pixels de large, 20% du debut avec critical bar, puis 80% restant avec bar :
	int crBarId = sprMngr.LoadSprite(criticalBar, SP_STATIC);
	int barId	= sprMngr.LoadSprite(bar, SP_STATIC);
	int w = sprMngr.GetW(crBarId);
	int spacing = w + w / 2;  // w * 1.5
	int numbars = 350 / spacing;
	int criticlimit = numbars / 5;  // * 20%

	if (etype == PB_STANDARD)
	{
		// nouvelle collection d'indices pour garder la trace des elements qu'on pousse :
		progressBars.push_back(vector<int>());
		for (int i = 0; i < numbars; ++i)
		{
			Element elem;
			elem.x = x + i * spacing;
			elem.y = y;
			if (i <= criticlimit)
				elem.spIds.push_back(crBarId);
			else
				elem.spIds.push_back(barId);
			// on ajoute un element de hud :
			elements.push_back(elem);
			// on ajoute l'indice dans cette progress bar :
			progressBars.back().push_back(elements.size() - 1);
		}
	}
	else if (etype == PB_WITH_ACTIVATED_PARTS)  // je l'avoue je n'avais pas prévu ce cas dès le départ
	{
		// on renseigne tout de meme un element dans le manager:
		Element e;
		e.x = x;
		e.y = y;
		e.spIds.push_back(-2);       // permettra d'identifer une PB_W.._A.._P..
		e.spIds.push_back(progressBars.size()); // indice pour acceder a la progressBar a partir des elements
		e.spIds.push_back(barId);    // indice du sprite de la barre en mode actif
		e.spIds.push_back(crBarId);  // indice du sprite inactif
		elements.push_back(e);       // ajout de l'element dans le manager

		// astuce a-la-con pour stocker les informations des progress bars avec parties activables
		progressBars.push_back(vector<int>(4));
		progressBars.back().at(0) = -2;                   // rend une progressbar speciale reconnaissable
		progressBars.back().at(1) = w + 1;                // stocke le spacing
		progressBars.back().at(2) = elements.size() - 1;  // indice de l'element correspondant
		progressBars.back().at(3) = 100;                  // valeur par defaut de la position (100%)
	}
	else
		throw std::invalid_argument("function Hud::PushProgressBar used with wrong enumeration");

	return progressBars.size() - 1;
}

void Hud::UpdateProgressBar(int barId, int progress)
{
	if (barId < 0 || barId >= (signed)progressBars.size())
		throw std::invalid_argument("function Hud::UpdateProgressBar called with an incorrect ID");

	bool isStandard = progressBars[barId][0] != -2;
	if (isStandard)
	{
		int numbars = progressBars[barId].size();
		int numvisible = (progress * numbars) / 100;
		for (int i = 0; i < numbars; ++i)
		{
			elements[ progressBars[barId][i] ].visible = (i <= numvisible) ? true : false;
		}
	}
	else  // progressbar a parties activables
	{
		progressBars[barId][3] = progress;
	}
}

void Hud::Draw(void)
{
	for (int i = 0; i < (signed)elements.size(); ++i)
	{
		if (elements[i].visible)
		{
			if (elements[i].numberFontID < 0)  // ce n'est pas un nombre
			{
				if (elements[i].spIds.size() == 4 && elements[i].spIds[0] == -2)  // c'est une progress bar speciale
				{
					DrawSpecialProgressBar(elements[i]);
				}
				else
				{  // element classique de HUD
					sprMngr.Draw(elements[i].spIds[ elements[i].current ], elements[i].x, elements[i].y);
				}
			}
			else
			{
				numbers[elements[i].numberFontID]->PrintNum(elements[i].x, elements[i].y, elements[i].value);
			}
		}
	}
	sprMngr.Update(UP_TIME_ONLY);
}

void Hud::DrawSpecialProgressBar(Element const & e)
{
	// dessinons une barre de 40 curseurs
	int x = e.x;
	int y = e.y;
	int spacing = progressBars[e.spIds[1]][1];
	for (size_t i = 0; i < 40; ++i)
	{
		if ((((signed)i*100)/40) < progressBars[e.spIds[1]][3])  // on est dans la partie active
			sprMngr.Draw(e.spIds[2], x + i * spacing, y);
		else
			sprMngr.Draw(e.spIds[3], x + i * spacing, y);
	}
}
