/**********************************************************************
 * File	: soundmanager.cpp
 * Author: Lightness1024
 *			  mailto:lightness1024@free.fr
 * Date	: 12/04/2008
 *********************************************************************/

#include "soundmanager.hpp"
#include <stdexcept>

// static :
int SoundManager::instances = 0;
SoundManager::EClosePolicy SoundManager::firstPolicy = CP_CLOSE_AT_LAST_DESTRUCT;
bool SoundManager::sdlInit = false;

// ------------------------------
// initialisation du manager
// ------------------------------
SoundManager::SoundManager(EClosePolicy policy /*= CP_CLOSE_AT_LAST_DESTRUCT*/)
{
	if (instances == 0)
	{
		int audio_rate = 44100;
		Uint16 audio_format = AUDIO_S16; // 16-bit stereo
		int audio_channels = 2;
		int audio_buffers = 2048;
		firstPolicy = policy;

		// lance l'audio dans SDL, une fois :
		if (!sdlInit && SDL_InitSubSystem(SDL_INIT_AUDIO) == -1)
		{
			fineinit = false;
			std::cout << "SDL_InitSubSystem returned an error" << std::endl;
		}
		else
		{
			//	lancement de la sous librairie haut niveau "SDL_mixer"
			if (Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers))
			{
				std::cout << "sound system didn't launch : " << Mix_GetError() << std::endl;
				fineinit = false;
			}
			else
			{
				Mix_AllocateChannels(64);  // au hasard, 16 types de sons differents.
				fineinit = true;	// biblio sonore correctement démarrée
			}
		}
		sdlInit = true;
		music = NULL;
	}
	++instances;
	if (instances > 1)  // en fait on ne peut pas s'en servir de plusieurs en même temps. trop risqué.
	{
		throw std::runtime_error("two sound managers can't be used on the same time");
	}
}


void SoundManager::PrepareMusicFile(const char* musfile)
{
	if (fineinit && music != 0)
	{
		StopMusic();
		Mix_FreeMusic(music);
	}
	music = Mix_LoadMUS(musfile);
}

void SoundManager::StartMusic(void)
{
	if (fineinit && music != 0 && !Mix_PlayingMusic())
		Mix_PlayMusic(music, -1);	 // en boucle
}

void SoundManager::StopMusic(void)
{
	if (fineinit && music != 0 && Mix_PlayingMusic())
		Mix_HaltMusic();
}

void SoundManager::SetMusicVolume(int volume)
{
	if (fineinit)
		Mix_VolumeMusic((volume * MIX_MAX_VOLUME) / 100);
}

bool SoundManager::Exist(const char* name)
{
	StrChunkMap::iterator it = chunks.find(name);  // chercher le son "name" dans la BDD
	return it != chunks.end();	  // si différent de end(), c'est qu'il existe.
}

void SoundManager::PrepareSoundFile(const char* sndfile, const char* name)
{
	if (fineinit)
	{
		if (!Exist(name))
			chunks[name] = Sample();	// création

		chunks[name].LoadSample(sndfile);  // charger le son demandé
	}
}

// ---- et puis les fonctions relais vers la classe "Sample" ----
//  retourne : le channel utilisé
//             ou
//             -1 si le manager de son n'a pas ete correctement initialisé
//             -2 si le son n'existe pas
//             -3 si le son n'a pas démarré a cause d'une trop forte attenuation due a la distance
int SoundManager::ReadSound(const char* name, int loops, bool newchannel, const Vector2i& soundLocation, bool forceStart /* = false*/)
{
	int usedChan = -1;
	if (fineinit)
	{
		usedChan = -2;
		StrChunkMap::iterator it = chunks.find(name);  // chercher
		if (it != chunks.end())  // s'il existe
		{
			usedChan = -3;
			float volume = 1.f;
			float fpan = 0.5f;

			// positionnement :
			if (soundLocation != Vector2i(0))  // lire un son localisé
			{
				std::pair< float, float > params = ComputeSoundPositionningParameters(soundLocation);
				volume = params.first;
				fpan = params.second;
			}
			if (forceStart || volume > 0.1f/*0.01f*/)    // on ne lance pas la lecture si la source est trop éloignée
			{
				usedChan = it->second.Play(loops, newchannel);
				it->second.SetVolume((int)(volume * 100));
				it->second.SetBalance((int)(fpan * 100));
			}
		}
	}
	return usedChan;
}

// fonction privée
// retourne une pair "volume, panning" au cas ou ca interesse quelqu'un
std::pair< float, float > SoundManager::ComputeSoundPositionningParameters(Vector2i const & position)
{
	float volume = 1.f;
	float fpan = 0.5f;

	// positionnement :
	// calculer le falloff et la balance en fonction de la position de l'écouteur
	int distance = (int)sqrtf((listenerPosition - position).LenghtSq());
	volume = std::max(MAX_LISTENING_DISTANCE - distance, 0) / (float)MAX_LISTENING_DISTANCE;
	fpan = position.x - listenerPosition.x + MAX_LISTENING_DISTANCE;
	fpan /= (float)MAX_LISTENING_DISTANCE * 2.f;
	fpan = std::max(fpan, 0.f);
	fpan = std::min(fpan, 1.f);

	return std::make_pair(volume, fpan);
}

// note : argument channel = -1 va faire les réglages sur le dernier channel utilisé
void SoundManager::UpdateSoundPosition(char const * name, Vector2i const & location, int channel /*= -1*/)
{
	if (fineinit)
	{
		StrChunkMap::iterator it = chunks.find(name);  // chercher "name"
		if (it != chunks.end())  // s'il existe
		{
			std::pair< float, float > ret = ComputeSoundPositionningParameters(location);
			// régler les paramètres trouvés ci-dessus 
			it->second.SetVolume((int)(ret.first * 100), channel);
			it->second.SetBalance((int)(ret.second * 100), channel);
		}
	}
}

void SoundManager::StopSound(const char* name, int channel)
{
	if (fineinit)
	{
		StrChunkMap::iterator it = chunks.find(name);  // comme fct "Read" ci-dessus
		if (it != chunks.end())
		{
			it->second.Stop(channel);
		}
	}
}

void SoundManager::SetSoundVolume(const char* name, int volume, int channel)
{
	if (fineinit)
	{
		StrChunkMap::iterator it = chunks.find(name);
		if (it != chunks.end())
		{
			it->second.SetVolume(volume, channel);
		}
	}	
}

void SoundManager::SetSoundBalance(const char* name, int balance, int channel)
{
	if (fineinit)
	{
		StrChunkMap::iterator it = chunks.find(name);
		if (it != chunks.end())
		{
			it->second.SetBalance(balance, channel);
		}
	}
}

void SoundManager::Update(const Vector2i& a_listenerPosition)
{
   listenerPosition = a_listenerPosition;
}

// static
void SoundManager::CloseSDLMIXAudio(void)
{
	Mix_CloseAudio();
}
		
SoundManager::~SoundManager()
{
	--instances;
	
	StopMusic();
	if (music != 0)
		Mix_FreeMusic(music);
	
	// désallouer tous les samples :
	chunks.clear();

	if (instances == 0)
	{
		if (firstPolicy == CP_CLOSE_AT_LAST_DESTRUCT)
		{
			CloseSDLMIXAudio();
		}
		firstPolicy = CP_CLOSE_AT_LAST_DESTRUCT;
	}
}

