/**********************************************************************
 * File	: soundmanager.hpp
 * Author: Lightness1024
 *			  mailto:lightness1024@free.fr
 * Date	: 11/04/2008
 *********************************************************************/

#ifndef SOUNDMANAGER_HPP_INC_19_03_2009___167344493_
#define SOUNDMANAGER_HPP_INC_19_03_2009___167344493_ 1

#include <SDL.h>
#include <SDL_mixer.h>

//	  STL
#include <map>
using std::map;
#include <iostream>
using std::cout;
using std::endl;
#include <string>
using std::string;

// ceci est l'en-tête du manager de sons pour extreme carnage

#include "geometry.hpp"


#define CHANNELS_PER_SAMPLE     32	      // channels par sample
#define MAX_LISTENING_DISTANCE  600       // 600 pixels


// ------------------------
// un son et ses actions
// ------------------------
class Sample
{
	Mix_Chunk* pchunk;
	int channel[CHANNELS_PER_SAMPLE];
	int lastchn;
	string fname;

public:

	void LoadSample(const char* file)
	{
		if (pchunk != 0)	  // si ya un chunk déjà chargé
			DeleteSample();
			
		pchunk = Mix_LoadWAV(file);  // WAVE, AIFF, RIFF, OGG, et VOC
			
		if (pchunk == 0)
			cout << "a sound failed to open : " << file << endl << " -> " << Mix_GetError() << endl;
		else
			fname = file;
	}
		
	// 0 = repetitions infinies, 1 = 1 fois, 2 = 2 ...
	// retourne le channel utilisé
	int Play(int loops = 1, bool newchannel = false)
	{
		if (pchunk != 0)	// si un son est alloué
		{
			if (newchannel)
			{
				++lastchn;
				lastchn %= CHANNELS_PER_SAMPLE;  // si l'user veut un nouveau channel, on tourne
			}
			if (IsPlaying(lastchn))
				Stop(lastchn);
			// lance la lecture pour un certain nombre de boucles.
			channel[lastchn] = Mix_PlayChannel(-1, pchunk, loops - 1);	// retourne le channel ou -1 si problème
		}
		return lastchn;
	}
		
	// -2 : arrete tous les channels
	// -1 : arrete le channel courant
	//	 i : arrete le channel spécifé
	void Stop(int a_chan = -1)
	{
		if (a_chan == -2)
		{
			for (int i = 0; i < CHANNELS_PER_SAMPLE; ++i)
				Stop(i);
			return;
		}
		else if (a_chan == -1)
			a_chan = lastchn;
		else
			a_chan %= CHANNELS_PER_SAMPLE;  // sécurité

		if (IsPlaying(a_chan))
		{
			Mix_HaltChannel(channel[a_chan]);	 // il n'est pas spécifié dans la doc de jcatki ce qui se passe si on appelle cette fonction
			// avec un channel qui ne joue pas ou qui n'est pas alloué -> d'où la sécurité IsPlaying()
			channel[a_chan] = -1;
		}
	}
		
	// retourne si ce sample est en train d'être joué ou non sur le channel "channel".
	// si channel = -1, il retourne si ca joue sur le channel courant
	//					 -2, sur un des channels au moins
	bool IsPlaying(int a_chan = -1)
	{
		// gestion des arguments spéciaux
		if (a_chan == -2)
		{
			for (int i = 0; i < CHANNELS_PER_SAMPLE; ++i)
			{
				if (IsPlaying(i))
					return true;
			}
			return false;
		}
		else if (a_chan == -1)
			a_chan = lastchn;
		else
			a_chan %= CHANNELS_PER_SAMPLE;
			
		// vraie fonction IsPlaying
		if (pchunk != 0 && Mix_GetChunk(channel[a_chan]) == pchunk)	 // si le son est alloué et son channel est actuellement sur ce son.
		{
			if (Mix_Playing(channel[a_chan]))  // si le channel est en lecture
			{
				return true;
			}
		}
		return false;
	}
		
	void SetVolume(int volume, int a_chan = -1)
	{
		if (a_chan == -1)
			a_chan = lastchn;

		if (IsPlaying(a_chan))
			Mix_Volume(channel[a_chan], volume * MIX_MAX_VOLUME / 100);		// règle le niveau du channel
		else if (pchunk != 0)
			Mix_VolumeChunk(pchunk, volume * MIX_MAX_VOLUME / 100);			// règle le niveau du chunk qui règlera le channel lors d'une prochaine lecture
	}
		
	void SetBalance(int balance, int a_chan = -1)
	{
		if (a_chan == -1)
			a_chan = lastchn;

		int left, right;
		left = 254 - balance * 254 / 100;
		right = balance * 254 / 100;

		if (IsPlaying(a_chan))
			Mix_SetPanning(channel[a_chan], left, right);
	}
		
	void DeleteSample(void)
	{
		Stop(-2);  // on arrête tous les channels

		if (pchunk != 0)
			Mix_FreeChunk(pchunk);		 // on libère la mémoire du chunk
		pchunk = 0;
		lastchn = 0;
	}

	// constructeur
	Sample(void)
	{
		for (int i = 0; i < CHANNELS_PER_SAMPLE; ++i)
			channel[i] = -1;
		lastchn = 0;
		pchunk = 0;
	}
		
	// constructeur de copie
	Sample(const Sample& src)
	{
		for (int i = 0; i < CHANNELS_PER_SAMPLE; ++i)
			channel[i] = -1;
		lastchn = 0;
		pchunk = 0;
		if (src.pchunk != 0)	  // si la source de la copie a un chunk de chargé
		{
			LoadSample(src.fname.c_str());
		}
	}

	// destructeur
	~Sample()
	{
		DeleteSample();
	}
};

// ------------------------
// manager de sons et musique
// ------------------------
class SoundManager
{
public:

	void PrepareMusicFile(char const * musfile);
	void StartMusic(void);
	void StopMusic(void);
	void SetMusicVolume(int volume);

	void Update(Vector2i const & listenerPosition);

	void PrepareSoundFile(char const * sndfile, char const * name);
	int  ReadSound(char const * name, int loops = 1, bool newchannel = false, Vector2i const & soundLocation = Vector2i(0), bool forceStart = false);
	void StopSound(char const * name, int channel = -1);
	void SetSoundVolume(char const * name, int volume = 100, int channel = -1);			// entre 0 et 100
	void SetSoundBalance(char const * name, int balance = 50, int channel = -1);		// 0 = gauche, 50 = milieu, 100 = droite
	void UpdateSoundPosition(char const * name, Vector2i const & location, int channel = -1);
	bool Exist(char const * name);
	
	static void CloseSDLMIXAudio(void);
	
	enum EClosePolicy
	{
		CP_CLOSE_AT_LAST_DESTRUCT,
		CP_DONT_CLOSE
	};
	SoundManager(EClosePolicy policy = CP_CLOSE_AT_LAST_DESTRUCT);
	~SoundManager();
	
private:
	struct AudioCannotStart{};	  // exception

	static int instances;  // compte les instances de cette classe
	static bool sdlInit;
	static EClosePolicy firstPolicy;

	Mix_Music* music;	  // channel réservé pour la musique

	typedef map< string, Sample > StrChunkMap;

	StrChunkMap chunks;	 // collection des samples (chunks)
		
	bool fineinit;

	Vector2i listenerPosition;
		
	std::pair< float, float > ComputeSoundPositionningParameters(Vector2i const & position);
};


#endif //SOUNDMANAGER_HPP_INC_19_03_2009___167344493_
