/**********************************************************************
 * File	: spritemanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 09/07/2007
 *********************************************************************/

#include "spritemanager.hpp"
#include "sharedsurface.hpp"
#include "myutils.hpp"
#include <stdexcept>

// définition des membres static

int SpriteManager::instances = 0;
render2d::SharedSurface* SpriteManager::screen_surface = NULL;
int SpriteManager::resx;
int SpriteManager::resy;
int SpriteManager::scroll_x = 0;
int SpriteManager::scroll_y = 0;
list< string > SpriteManager::windowTitles;

// bliter un sprite
void SpriteManager::Draw(int id, int posX, int posY, float frametime)
{
	Sprite& sprite = *(ap_sprites[id]);
	if (posX != -1)
	{
		sprite.posx = posX;
		sprite.posy = posY;
	}

	SDL_Rect rcSrc, rcDest;

	if (sprite.nbFrames > 0)  // sprite à animation
	{
		if (frametime >= 0.f)  // l'user spécifie l'animation à la main
		{
			sprite.infos.animation = frametime * sprite.speed;
		}
		else
		{
			// pour animer, on avance dans les frames:
			if (sprite.lastUpdateFrame != m_frame)	 // si ce sprite n'a jamais encore été dessiné à cette frame
			{
				double tv = globalTime - lastTiming;             // interval de temps pour l'animation
				sprite.infos.animation += tv * sprite.speed;	// on augmente l'animation
				sprite.lastUpdateFrame = m_frame;
			}
		}
		int frame = myutils::abs((int)sprite.infos.animation);
		if (frametime >= 0.f && frame >= sprite.nbFrames)	// en animation manuelle, on ne boucle pas.
		{
			rcSrc.x = 0;
			rcSrc.y = 0;
			rcSrc.w = 0;
			rcSrc.h = 0;
		}
		else
		{
			frame %= sprite.nbFrames;
			rcSrc.x = frame * sprite.surface->h;  // frame * largeur du carre = frame * hauteur de l'image
			rcSrc.y = 0;
			rcSrc.w = sprite.surface->h;			  // w = hauteur (car le sprite est carre)
			rcSrc.h = sprite.surface->h;
		}
	}
	else if (sprite.nbFrames == 0)  // sprite a rotation
	{
		// l'angle du sprite n'est jamais plus eloigne du vrai angle que de "delta" / 2.
		float angle = sprite.infos.angles.angle + sprite.infos.angles.angle_delta * 0.5f;
		int w = sprite.surface->w >> 2;	// divise par 4
		// determiner la case a lire dans le sprite selon l'angle
		if (angle < 180.f)  // premier cas, descendant.
		{
			if (angle < 90)
			{
				rcSrc.x = 0;
				rcSrc.y = (int)(angle / sprite.infos.angles.angle_delta);
			}
			else
			{
				rcSrc.y = (int)((angle - 90) / sprite.infos.angles.angle_delta);
				rcSrc.x = w;
			}
			rcSrc.y *= w;
		}
		else	// au dela de 180, c'est montant.
		{
			if (angle < 270)
			{
				rcSrc.x = w << 1;	 // fois 2
				rcSrc.y = sprite.surface->h / w - 1 - (int)((angle - 180) / sprite.infos.angles.angle_delta);
			}
			else
			{
				rcSrc.y = sprite.surface->h / w - 1 - (int)((angle - 270) / sprite.infos.angles.angle_delta);
				if (rcSrc.y < 0)
					rcSrc.y = 0;	// possible car angle peut aller jusqu'a 360 + d/2
				rcSrc.x = w * 3;
			}
			rcSrc.y *= w;
		}
		rcSrc.w = w;
		rcSrc.h = w;
	}
	else	// sprite statique
	{
		rcSrc.x = 0;
		rcSrc.y = 0;
		rcSrc.w = sprite.surface->w;
		rcSrc.h = sprite.surface->h;
	}

	if (scroll_sensible)	 // tenir compte du scrolling de l'ecran.
	{
		rcDest.x = sprite.posx - scroll_x;	 // on pose le sprite a la position "coin en haut a gauche" du sprite.
		rcDest.y = sprite.posy - scroll_y;
	}
	else
	{
		rcDest.x = sprite.posx;
		rcDest.y = sprite.posy;
	}

	// w et h ne sont pas lus.

	SDL_BlitSurface(ap_sprites[id]->surface,
					&rcSrc,
					screen_surface->SdlPtr(),
					&rcDest);
}

// charger un sprite
int SpriteManager::LoadSprite(const char* fname, SPRITE_PARAM_en parm)
{
	// ajouter un sprite:
	ap_sprites.Add(smart_ptr<Sprite>(new Sprite));

	MapNameSurface::iterator it;
	it = surfaces_names.find(string(fname));
	if (it != surfaces_names.end())	// ce fichier a deja ete charge
	{
		ap_sprites.Back()->surface = it->second;	// on reference la surface existante
	}
	else
	{
		SDL_Surface* ptemp = SDL_LoadBMP(fname);	// sinon on la charge
		if (ptemp == NULL)
		{
			throw std::runtime_error(string(fname) + " - no such file/load error");
		}
		else
		{
			ap_sprites.Back()->surface = SDL_DisplayFormat(ptemp);
			SDL_FreeSurface(ptemp);
			surfaces_names[fname] = ap_sprites.Back()->surface;
			int colorkey = SDL_MapRGB(ap_sprites.Back()->surface->format, 255, 0, 255);
			SDL_SetColorKey(ap_sprites.Back()->surface, SDL_SRCCOLORKEY | SDL_RLEACCEL, colorkey);
		}
	}
	ap_sprites.Back()->posx = 0;
	ap_sprites.Back()->posy = 0;
	ap_sprites.Back()->speed = 4.0f;
	ap_sprites.Back()->lastUpdateFrame = m_frame - 1;

	if (parm == SP_ROTATED)
	{
		ap_sprites.Back()->nbFrames = 0;
		ap_sprites.Back()->infos.angles.angle_delta = 360.f / (16 * ap_sprites.Back()->surface->h / ap_sprites.Back()->surface->w);
		ap_sprites.Back()->infos.angles.angle = 0.f;
	}
	else if (parm == SP_ANIMATED)
	{
		ap_sprites.Back()->infos.animation = 0.0;
		ap_sprites.Back()->nbFrames = ap_sprites.Back()->surface->w / ap_sprites.Back()->surface->h;
	}
	else if (parm == SP_STATIC)
	{
		ap_sprites.Back()->nbFrames = -1;
	}

	return ap_sprites.GetUbound();
}

void SpriteManager::Update(UPDATE_PARAM_en opt /*= UP_SCREEN_AND_TIME*/)
{
	// internal timer usage (automatic animation time management)
	lastTiming = globalTime;
	globalTime = internalTimer.GetTime();
	if (opt == UP_SCREEN_AND_TIME)
	{
		SDL_Flip(screen_surface->SdlPtr());
	}
	++m_frame;
}


void SpriteManager::UpdateWithClientTime(double globalTime_arg, UPDATE_PARAM_en opt /*= UP_SCREEN_AND_TIME*/)
{
	// client-driven animation time
	if (lastTiming - 1e-5 > globalTime_arg)
		throw std::runtime_error("time is going backwards");

	lastTiming = globalTime;
	globalTime = globalTime_arg;
	if (opt == UP_SCREEN_AND_TIME)
	{
		SDL_Flip(screen_surface->SdlPtr());
	}
	++m_frame;
}

void SpriteManager::ClearScreen(int color /*= 0*/)
{
	SDL_FillRect(screen_surface->SdlPtr(), NULL, color);
}

// initialiseur
void SpriteManager::Init(int resolutionX, int resolutionY, const char* title)
{
	if (!instances && screen_surface)
	{
		throw std::runtime_error("no screen surface should exist");
	}
	else if (instances && !screen_surface)
	{
		throw std::runtime_error("a screen surface should exist");
	}

	if (instances == 0 && !render2d::SharedSurface::Exists())
	{
		screen_surface = new render2d::SharedSurface(resolutionX, resolutionY, render2d::DP_QUIT_SDL);
		resx = resolutionX;
		resy = resolutionY;
	}
	else if (instances == 0 && render2d::SharedSurface::Exists())
	{
		screen_surface = new render2d::SharedSurface();  // prend l'existante en reference
		resx = screen_surface->GetX();
		resy = screen_surface->GetY();
	}

	std::string finaltitle = title;
	if (title == std::string(""))
		finaltitle = windowTitles.empty() ? "no name" : windowTitles.back();

	SDL_WM_SetCaption(finaltitle.c_str(), finaltitle.c_str());
	windowTitles.push_back(finaltitle);

	scroll_sensible = true;
	internalTimer.Start();
	lastTiming = 0.;
	globalTime = 0.;
	++instances;
	ap_sprites.ReserveExtra(20);
	m_frame = 0;
}

SpriteManager::~SpriteManager()
{
	MapNameSurface::iterator it;
	for (it = surfaces_names.begin(); it != surfaces_names.end(); ++it)
	{
		SDL_FreeSurface(it->second);	 // liberation des surfaces
	}
	--instances;
	if (instances > 0)
	{
		// restauration de l'ancien nom de la fenêtre
		windowTitles.pop_back();
		SDL_WM_SetCaption(windowTitles.back().c_str(), windowTitles.back().c_str());
	}
	else   // intances <= 0
	{
		delete screen_surface;
		screen_surface = NULL;
		resx = 0;
		resy = 0;
		scroll_x = 0;
		scroll_y = 0;
	}
}
