/**********************************************************************
 * File	: spritemanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 09/07/2007
 *********************************************************************/

#ifndef SPRITEMANAGER_HPP_INC_19_03_2009__107065064_
#define SPRITEMANAGER_HPP_INC_19_03_2009__107065064_ 1

#include <map>
#include <string>
#include <list>
#include <SDL/SDL.h>

#include "SecureArray.hpp"
#include "timer.hpp"
#include "smart_ptr.hpp"
#include "geometry.hpp"

#include "noncopyable.hpp"

using std::map;
using std::string;
using std::list;

struct Sprite
{
	int posx, posy;
	SDL_Surface* surface;
	int nbFrames;
	union Infos
	{
		double animation;
		struct SprAngle
		{
			float angle;
			float angle_delta;
		} angles;
	} infos;
	float speed;
	int lastUpdateFrame;
};

typedef smart_ptr< Sprite > SpriteSmartPtr;

enum UPDATE_PARAM_en
{
	UP_SCREEN_AND_TIME,
	UP_TIME_ONLY
};

enum SPRITE_PARAM_en
{
	SP_STATIC,
	SP_ANIMATED,
	SP_ROTATED
};

namespace render2d
{
	class SharedSurface;
}

class SpriteManager : boost::noncopyable
{
	typedef map< string, SDL_Surface* > MapNameSurface;
		
	MapNameSurface surfaces_names;
	SecureArray< SpriteSmartPtr > ap_sprites;
	static list< string > windowTitles;
	tools::HiResTimer internalTimer;
	double globalTime;
	double lastTiming;
	bool scroll_sensible;
	int m_frame;
		
	static render2d::SharedSurface* screen_surface;
		
	static int instances;
	static int resx, resy;

public:
		
	static int scroll_x, scroll_y;  // shared
	
	int LoadSprite(char const* fname, SPRITE_PARAM_en parm = SP_ANIMATED);
	void Draw(int id, int posX = -1, int posY = -1, float frametime = -1.f);
	void DrawAll(unsigned id_from = 0)
	{
		for (int i = id_from; i < ap_sprites.GetElementsNumber(); ++i)
		{
			Draw(i);
		}
	}
	int GetNumSprites(void)
	{
		return ap_sprites.GetElementsNumber();
	}
	void Move(int id, int posX, int posY)
	{
		ap_sprites[id]->posx = posX;
		ap_sprites[id]->posy = posY;
	}
	void MoveRelative(int id, int displaceX, int displaceY)
	{
		ap_sprites[id]->posx += displaceX;
		ap_sprites[id]->posy += displaceY;
	}
	void SetSpeed(int id, float speed)
	{
		ap_sprites[id]->speed = speed;
	}
	void Rotate(int id, float angle)
	{
		angle = FltModulo(angle, 360.f);
		ap_sprites[id]->infos.angles.angle = angle;
	}
	int GetW(int id) const
	{
		if (ap_sprites[id]->nbFrames > 0)	 // anime
			return ap_sprites[id]->surface->h;	// square
		else if (ap_sprites[id]->nbFrames == 0)
			return ap_sprites[id]->surface->w >> 2;  // a rotation, 4 cadrans en colonne
		else
			return ap_sprites[id]->surface->w;
	}
	int GetH(int id) const
	{
		if (ap_sprites[id]->nbFrames < 0)
			return ap_sprites[id]->surface->h;
		else
			return GetW(id);	 // anime ou rotate, les sprites sont carres.
	}
	static int GetResX(void)
	{
		return resx;
	}
	static int GetResY(void)
	{
		return resy;
	}
	void SetScrollSensible(bool ss)
	{
		scroll_sensible = ss;
	}
	
	void Update(UPDATE_PARAM_en opt = UP_SCREEN_AND_TIME);
	void UpdateWithClientTime(double global_time, UPDATE_PARAM_en opt = UP_SCREEN_AND_TIME);
		
	void ClearScreen(int color = 0);
	
	// constructeurs
	SpriteManager(int resolutionX, int resolutionY, char const* windowTitle)
	{
		Init(resolutionX, resolutionY, windowTitle);
	}
	SpriteManager(char const* windowTitle)
	{
		Init(-1, -1, windowTitle);
	}
	SpriteManager(void)
	{
		Init(640, 480, "");
	}

	~SpriteManager();

private:

	void Init(int resolutionX, int resolutionY, char const* windowTitle);
};

// cette fonction permet a un appelant de trouver la coordonnée de départ en X pour
// un sprite qu'il devrait centrer au milieu de l'écran.
// exemple : affichage de menu
inline int GetScreenCenterX(int spriteWidth)
{
	return (SpriteManager::GetResX() - spriteWidth) / 2;
}

inline int GetScreenCenterY(int spriteHeight)
{
	return (SpriteManager::GetResY() - spriteHeight) / 2;
}


#endif //SPRITEMANAGER_HPP_INC_19_03_2009__107065064_
