/**********************************************************************
 * File	: terraingenerator.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 10/07/2007
 *********************************************************************/

#ifndef TERRAINGENERATOR_HPP_INC_19_03_2009__87038113_
#define TERRAINGENERATOR_HPP_INC_19_03_2009__87038113_ 1

#include "geometry.hpp"
#include "randclass.hpp"
#include "SecureArray.hpp"
#include "spritemanager.hpp"
#include <list>

using std::list;

typedef list< const char * > CharPtrList;

// enumerate all kind of object we can have in a map's setup :
// from bigger to smaller is better for procedural generation
enum KIND_OF_TERRAIN_OBJECTS_en
{
	KTO_WATER,
	KTO_ROCKS,
	KTO_TREE,
	
	NB_KIND_OF_TERRAIN_OBJECTS
};

struct TerrainParameters
{
	int seed;
		
	// unit is / l'unite est :
	// nombre de repetitions du sprite de terrain
	// number of tiling repeats of background sprite
	int size_long;
	int size_height;
		
	// unit is :
	// 0 = no ; 1.0 = full coverage
	float density[NB_KIND_OF_TERRAIN_OBJECTS];
		
	// file name for ground tile :
	const char* terrain_sprite;
		
	// file name for external border wall sprite :
	const char* wall_sprite;
		
	// file name for final line sprite :
	const char* finalline_sprite;

	// list of sprite's file name strings for each kind of terrain objects:
	CharPtrList sprites[NB_KIND_OF_TERRAIN_OBJECTS];
};

struct ORect;

struct TerrainObject
{
	Rect rect;
	int spriteId;
	bool highplane;
		
	TerrainObject(const Rect& a_rc, int spId, bool highZ)
			: rect(a_rc),
			  spriteId(spId),
			  highplane(highZ) { }

	virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool probe = false) = 0;
	virtual bool IntersectORect(const ORect& orc) = 0;
	virtual bool IntersectRect(const Rect& rc) = 0;

	virtual ~TerrainObject() { }
};

struct Tree : TerrainObject
{
	Tree(const Rect& a_rc, int spId)
			: TerrainObject(a_rc, spId, true) { }
		
	virtual bool IntersectRect(const Rect& rc)
	{
		return ORect(rc).IntersectsCircle(rect.GetCenter(), rect.w >> 3);
	}
	virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool /*probe*/)
	{
		return CircleRayCollides(t, start, direction, rect.GetCenter(), rect.w >> 3);
	}
	virtual bool IntersectORect(const ORect& a_orc)
	{
		return a_orc.IntersectsCircle(rect.GetCenter(), rect.w >> 3);
	}
};

struct Water : TerrainObject
{
	Water(const Rect& a_rc, int spId)
			: TerrainObject(a_rc, spId, false) { }
		
	virtual bool IntersectRect(const Rect& rc)
	{
		return ORect(rc).IntersectsCircle(rect.GetCenter(), rect.w >> 1);
	}
	virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool probe = false)
	{
		if (probe)			 // les rayons de sondage doivent etre interceptés pour que l'IA evite les étangs.
			return CircleRayCollides(t, start, direction, rect.GetCenter(), rect.w >> 1);
		else
			return false;	 // mais on peut tirer au dessus de l'eau
	}
	virtual bool IntersectORect(const ORect& a_orc)
	{
		return a_orc.IntersectsCircle(rect.GetCenter(), rect.w >> 1);
	}
};

struct Rocks : TerrainObject
{
	Rocks(const Rect& a_rc, int spId)
			: TerrainObject(a_rc, spId, false) { }
		
	virtual bool IntersectRect(const Rect& rc)
	{
		return rect.Intersects(rc);
	}
	virtual bool IntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool /*probe*/)
	{
		ORect orc(rect);
		return orc.IntersectsRay(t, start, direction);
	}
	virtual bool IntersectORect(const ORect& a_orc)
	{
		return ORect(rect).IntersectsORc(a_orc);
	}
};

struct FinalLine : TerrainObject
{
	FinalLine(const Rect& a_rc, int spId)
			: TerrainObject(a_rc, spId, false) { }
		
	virtual bool IntersectRect(const Rect& /*rc*/)
	{
		return false;
	}
	virtual bool IntersectRay(float* /*t*/, const Vector2f& /*start*/, const Vector2f& /*direction*/, bool /*probe*/)
	{
		return false;
	}
	virtual bool IntersectORect(const ORect& /*a_orc*/)
	{
		return false;
	}
};

struct ObjInstancier
{
	virtual TerrainObject* Create(const Rect&, int) = 0;
	virtual ~ObjInstancier() { }
};
		
template< class C >
struct Instancier : ObjInstancier
{
	C* Create(const Rect& rc, int id)
	{
		return new C(rc, id);
	}
};

template< typename Data_T >
struct BiNode
{
	BiNode* left;
	BiNode* right;

	Data_T data;
		
	bool IsLeaf(void) const
	{
		return left == NULL && right == NULL;
	}
		
	BiNode(void)
			: left(NULL),
			  right(NULL) { }
		
	BiNode(const Data_T& a_dt)
			: left(NULL),
			  right(NULL),
			  data(a_dt) { }

	~BiNode()
	{
		if (left)
			delete left;
		if (right)
			delete right;
	}
};

struct Sector
{
	struct FlagIndex
	{
		int index;				// reference to object stored at "index" in array "objects" of TM class.
		int flag;
		FlagIndex(void)
				: flag(-1) { }

		FlagIndex(int i, int f = -1)
				: index(i),
				  flag(f) { }
	};
	SecureArray< FlagIndex > contents;
	Rect rect;		
};

typedef BiNode< Sector > SectorNode;

class TerrainManager
{
	// space partitionning :
	SectorNode root;
	// scene container :
	SecureArray< TerrainObject* > objects;
		
	// water, terrain, rocks
	SpriteManager background;
	// les arbres doivent etre dessinnes au dessus des voitures:
	// trees must be drawn above cars:
	SpriteManager highground;
		
	bool loaded;
	Vector2i levelSize;
	int borderWidth;
		
	// tableau persistant pour stocker des noeuds actifs
	SecureArray< SectorNode* > m_nodeList;
	int m_numNodes;
	int m_frame;
		
	static int instances;
		
	// -- private methods --
	void BuildInDepth(SectorNode& node);

	CharPtrList::const_iterator GetRandomIt(const CharPtrList& a_list, int max, tools::Random& rand) const;

	void AddFromNode(int index, SectorNode& node);
	inline void InsertObject(TerrainObject* pobj);
	void BuildActivatedNodesList(const Rect& rc, SectorNode& node, SecureArray< SectorNode* >& nodeList, int& numNodes);
		
	enum OBJ_INTERSECT_TYPE_en
	{
		OIT_BOUNDING_RECT,
		OIT_RECT,
		OIT_ORECT
	};
		
	bool IsOverObjectFromNode(const ORect& orc, const Rect& bounding, const SectorNode& node, OBJ_INTERSECT_TYPE_en type);
	bool IsTypeIntersectingObjectOfNode(const ORect& orc, const SectorNode& node, OBJ_INTERSECT_TYPE_en type);

public:
		
	// -- public interface --
		
	void DrawBackGround(void);
	void DrawHighGround(void);

	void Update(double globalTime, UPDATE_PARAM_en p = UP_TIME_ONLY)
	{
		background.UpdateWithClientTime(globalTime, p);
		highground.UpdateWithClientTime(globalTime, p);
	}
	void Scroll(int x, int y)
	{
		SpriteManager::scroll_x = x;
		SpriteManager::scroll_y = y;
	}
	Vector2i GetSize(void)
	{
		return levelSize;
	}
	int GetBorderWidth(void)
	{
		return borderWidth;
	}
		
	void Generate(TerrainParameters& params);
		
	// collisions detection functions :
	bool IsOverObject(const Rect& rc)
	{
		return IsOverObjectFromNode(rc, rc, root, OIT_BOUNDING_RECT);
	}
	bool WorldIntersectRect(const Rect& rc)
	{
		return IsOverObjectFromNode(rc, rc, root, OIT_RECT);
	}
	bool WorldIntersectORect(const ORect& a_orc)
	{
		return IsOverObjectFromNode(a_orc, a_orc.GetBoundingRect(), root, OIT_ORECT);
	}
	bool WorldIntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool probe = false, bool lowspan = true);
		
	struct TooManyInstances {};
		
	TerrainManager(void)
			: loaded(false),
			  levelSize(0, 0),
			  borderWidth(0),
			  m_frame(0)
	{
		if (++instances != 1)
			throw TooManyInstances();
		m_nodeList.Resize(32);
	}

	// destroy managed objects when time has come to die :
	~TerrainManager()
	{
		for (int i = 0; i < objects.GetElementsNumber(); ++i)
		{
			delete objects[i];
		}
		--instances;
	}
};

void TerrainManager::InsertObject(TerrainObject* pobj)
{
	// ajout a la collection :
	objects.Add(pobj);
	// puis ajout de l'indice dans les listes des feuilles qui intersectent :
	// add new index in leaves's lists that intersects :
	AddFromNode(objects.GetUbound(), root);
}

#endif //TERRAINGENERATOR_HPP_INC_19_03_2009__87038113_
