/**********************************************************************
 * File	: towermanager.cpp
 * Author: Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 01/08/2007
 *********************************************************************/

#include "towermanager.hpp"
#include "terraingenerator.hpp"
#include "collisionmanager.hpp"
#include "carsmanager.hpp"
#include "supramanager.hpp"

int TowerManager::instances = 0;

#define FIRE_RANGE 400

static inline Vector2i& VBitshift(Vector2i& in, int bits)
{
	in.x >>= bits;
	in.y >>= bits;
	return in;
	// le compilateur vas changer ca tout seul en une jolie forme RVO-compliant :
	// "return VBitshift(in.x >> bits, in.y >> bits);"
}

bool TowerManager::TowerIntersectProbeRay(float* t, const Vector2f& start, const Vector2f& direction)
{
	// on active quatre cellules
	Vector2i pt1 = start;	// une sur la position de depart
	Vector2i pt2, pt3, pt4;		  // et les autres autours

	ClIt clls[4];

	Vector2i dir5px = direction * 5;

	pt2 = pt1 + dir5px;
	pt3 = pt2 + dir5px;
	pt4 = pt3 + dir5px;

	clls[0] = cells.find( VBitshift(pt1, log2cellsize) );
	clls[1] = cells.find( VBitshift(pt2, log2cellsize) );
	clls[2] = cells.find( VBitshift(pt3, log2cellsize) );
	clls[3] = cells.find( VBitshift(pt4, log2cellsize) );
	float tt = 1e37f;
	float tmin = 1e37f;
	for (int i = 0; i < 4; ++i)	// pour les quatres cellules
	{
		if (clls[i] != cells.end())	// si elles existent
		{
			TSPCell::TwIt towerIt = clls[i]->second->towers.begin();
			for (; towerIt != clls[i]->second->towers.end(); ++towerIt) // pour toute tour dans celle cellule
			{
				if (ORect(towerIt->rect).IntersectsRay(&tt, start, direction))
				{
					if (tt < tmin)
						tmin = tt;	// on garde le t minimal
				}
			}
		}
	}
	if (tmin < 1e36f)
	{
		*t = tmin;
		return true;
	}
	else
		return false;
}

bool TowerManager::TowerIntersectORect(const ORect& orc, bool bounding)
{
	// d'abord on augmente les côtés haut et gauche du rectangle de maxw,maxh pour déterminer les cellules activées:
	// car les tourelles sont stockées de manière unique (dans UNE SEULE cellule). En fonction de leur coin haut-gauche.
	// donc si une tourelle déborde dans la cellule de droite ou du bas, ou aussi du coin bas-droit,
	// il faut activer aussi les cellules susceptibles de contenir les potentielles tourelles qui débordent.

	Vector2i corners[4];
	corners[0] = orc.rect.pos - Vector2i(maxw, maxh);
	corners[2] = orc.rect.pos + Vector2i(orc.rect.w, orc.rect.h);
	// de temps en temps on aimerais bien utiliser SSE
	corners[0].x >>= log2cellsize;
	corners[0].y >>= log2cellsize;
	corners[2].x >>= log2cellsize;
	corners[2].y >>= log2cellsize;
	corners[1].x = corners[2].x;
	corners[1].y = corners[0].y;
	corners[3].x = corners[0].x;
	corners[3].y = corners[2].y;
	// les cellules étant nécessairement plus grandes (du moins j'espère) que les ennemis augmentés,
	// on ne peut activer que 4 cellules au maximum.
	// (c'est exactement le meme problème que la grille des hostiles, voir hostilesmanager.{h,c}pp)

	ClIt cellItors[5];
	cellItors[0] = cells.end();  // 1 pour contenir end()
	for (int i = 1; i < 5; ++i)
	{
		// recherche dans la map
		// (oui les cellules sont contenues dans une map et spawnées dynamiquement en fonction des besoins) :
		cellItors[i] = cells.find(corners[i - 1]);	// cellule activée par le coin 'i'
		bool new1 = true;
		for (int j = 0; j < i; ++j)  // de 0 a i.
		{
			if (cellItors[j] == cellItors[i])
			{
				new1 = false;	// on a déjà rencontré cet itérateur. (ou cellItors[i] == end())
				break;
			}
		}
		if (new1)	// seulement si cellItors[i] est un itérateur encore non-traité.
		{	// oui car on ne souhaite pas traiter deux fois la même cellule.
			// verifier la collision entre orc et toutes les tourelles de cette cellule:
			TSPCell::TwIt towerIt = cellItors[i]->second->towers.begin();
			for (; towerIt != cellItors[i]->second->towers.end(); ++towerIt)
			{
				if (bounding)
				{
					if (orc.rect.Intersects(towerIt->rect))
						return true;
				}
				else
				{
					if (orc.IntersectsORc(towerIt->rect))
						return true;
				}
			}
		}
	}
	return false;
}

// instancier une tourelle
bool TowerManager::BuildNewTower(int x, int y, int kindOfTower, Gunner* builder)
{
	// vérifier si l'id donné existe :
	if (towerKinds.GetUbound() < kindOfTower)
		return false;	// spo possible de construire un truc qu'on sait pas quoi que c'est. gné

	const int twrw = sprMgr.GetW(towerKinds[kindOfTower].spriteId);
	const int twrh = sprMgr.GetH(towerKinds[kindOfTower].spriteId);
	const int cx = x;
	const int cy = y;
	x -= twrw >> 1;
	y -= twrh >> 1;

	// vérifier si on peut construire la tourelle à cette position :
	Rect twrrc = Rect(x, y,
					  twrw,
					  twrh);
	if (ChiefManager::GetCollisionsAndProbing()->IntersectRect(twrrc))
		return false;

	// vérifier la presence d'ennemis dans la zone de la tourelle et les detruire
	list< Car* > crushedCars;
	ChiefManager::GetHostilesManager()->HostileIntersectRect(crushedCars, twrrc);

	list< Car* >::iterator ccit;
	// boucler la liste (remplie par reference par la fonction HostileIntersectRect)
	// si aucun ennemi dans la liste, la boucle sera sautee.
	for (ccit = crushedCars.begin(); ccit != crushedCars.end(); ++ccit)
	{
		// on détruit les hostiles écrasés en ramenant leurs points de vie a 0. personne ne beneficie de leur mort.
		// (ce serait trop facile de tuer des ennemis en posant des tourelles :3 )
		(*ccit)->health = 0;	 // boum
	}

	// déterminer la position dans la grille :
	x >>= log2cellsize;
	y >>= log2cellsize;
	Vector2i gridpos(x, y);
	ClIt cell = cells.find(gridpos);
	if (cell == cells.end())  // il n'existe pas encore de cellule à cette position
	{
		// insérons la :
		cell = cells.insert(ClValT(gridpos, smart_ptr<TSPCell>(new TSPCell))).first;
	}
	// ajouter une tourelle a la freelist de cette cellule (towers) :
	TowerInstance newtower(kindOfTower, builder);
	// quelques renseignements en plus:
	newtower.rect = twrrc;
	newtower.GiveWeapon(towerKinds[kindOfTower].weaponId);	// on donne une arme a la tourelle
	// demander au manager d'hostiles de construire le firerange autours de cette tourelle
	newtower.firingRange = ChiefManager::GetHostilesManager()->BuildFireRangeCellList(Vector2i(cx, cy), FIRE_RANGE);
	// son ia class copiée du builder:
	newtower.SetIAClass(builder->GetIAClassName().c_str(), ChiefManager::GetIAManager());
	// vive les smart pointers parce que avec toutes ces copies...
	cell->second->towers.push_back(newtower);

	return true;
}

// renseigner la base de donnée d'un type de tourelle
int TowerManager::CreateNewKindOfTower(const char* sprite, int weaponId, int price)
{
	int id = sprMgr.LoadSprite(sprite);
	towerKinds.Add(Tower(id, weaponId, price));
	int w = sprMgr.GetW(id);
	int h = sprMgr.GetW(id);
	if (w > maxw)
		maxw = w;
	if (h > maxh)
		maxh = h;
	return towerKinds.GetUbound();
}

// mise a jour : IA, affichage
void TowerManager::Update(double gtime)
{
	static int s_mod8shifting;

	// on itère sur toutes les tourelles pour mettre à jour leur IA
	for (ClIt cellit = cells.begin(); cellit != cells.end(); ++cellit)	// pour toutes les cellules
	{
		// et pour toutes les tourelles dans chaque cellule:
		TSPCell::TwIt twrit = cellit->second->towers.begin();
		int counter = 0;
		for (; twrit != cellit->second->towers.end();  // while
			  ++twrit, ++counter)					   // steps
		{
			// on dessine la tourelle :
			sprMgr.Draw(towerKinds[ twrit->towerKind ].spriteId, twrit->rect.pos.x, twrit->rect.pos.y);

			// pour des raisons d'optimisations, on ne rafraichi l'IA que d'une tourelle sur 8
			// tourelles no 0, 8, 16.. puis a l'image d'apres on fera les tourelles 1, 9, 17 etc.

			if (counter % 8 == s_mod8shifting)	 // wrapper de rafraichissement modulo 8
			{
				UpdateIAOnTower(twrit);
			} // fin wrapper de refresh
		}
	} // fin "pour toutes les cellules" et "pour toutes les tourelles"
	sprMgr.UpdateWithClientTime(gtime, UP_TIME_ONLY);
	++s_mod8shifting %= 8;	// 0,1,2,...,6,7,0,1,2...
}

void TowerManager::UpdateIAOnTower(TSPCell::TwIt& tower)
{
	// il est possible de perdre le tracking d'un ennemi pour les raisons suivantes :
	//	 - nous l'avons detruit
	//	 - il a depasse la fin du niveau, et disparu
	//	 - un autre gunner l'a detruit
	// si cela arrive, notre cible est automatiquement passee a NULL par le destructeur (C++) de l'ennemi lui meme
	if (tower->currentTarget != 0)  // si on traque actuellement un ennemi
	{
		// verifier qu'il est toujours a portee et que ca a un interet de tirer sur lui:
		UpdateTargetReachability(tower);
	}

	// on est OK, on continue de tirer sur notre cible actuellement traquee, ou bien,
	if (tower->currentTarget == 0)  // nous n'avons plus de traque
	{
		// il faut alors demander au manager des hostiles de nous trouver un hostile a portee :
		FindTarget(tower);
	}

	// si le manager d'hostiles nous a trouve une nouvelle cible (ou qu'on reste sur la precedente) :
	if (tower->currentTarget)
	{
		// rafraichir la direction et l'angle de tir :
		RefreshFireDir(tower);
	}
	else // traque perdue, et pas de nouvel ennemi a portee.
	{
		// interrompre le tir
		ChiefManager::GetWeaponManager()->CeaseFire(&(*tower));
		// rappel : la notation &(*x) signifie l'adresse d'un objet dont l'operateur * sur x renvoi la reference (x n'est cependant PAS un pointeur).
	}

	// et de temps en temps aussi :
	tower->TransferKills(tower->builder);   // crédite tous les kills effectués par la tourelle a son constructeur.
}

void TowerManager::UpdateTargetReachability(TSPCell::TwIt& tower)
{
	Vector2f fireRay = tower->currentTarget->rect.GetCenter() - tower->rect.GetCenter();
	float distance = sqrt(fireRay.LenghtSq());
	if (distance > FIRE_RANGE || tower->currentTarget->health <= 0)	 // trop loin ou déjà mort (sans d'interet de tirer dessus)
	{
		// traque perdue !
		tower->currentTarget->UnregisterTracker(&(*tower));
		tower->currentTarget = 0;
	}
	else
	{
		float t = 1e36f;
		Vector2f dir = fireRay * (1.f / distance);
		// verifier la liberte de la ligne de mire :
		if (ChiefManager::GetCollisionsAndProbing()->IntersectRay(&t, tower->rect.GetCenter(), dir, false, false))
		{
			if (t < distance)	 // quelque chose se trouve entre la tourelle et notre cible
			{
				// traque perdue :'(
				tower->currentTarget->UnregisterTracker(&(*tower));
				tower->currentTarget = 0;
			}
		}
	}
}

void TowerManager::FindTarget(TSPCell::TwIt& tower)
{
	tower->currentTarget = ChiefManager::GetHostilesManager()->GiveMeAnEnnemiInRange(tower->firingRange, *tower);
	if (tower->currentTarget)	// si trouve qqch
	{
		// alors on s'enregistre comme tracker actif aupres de cet ennemi:
		tower->currentTarget->RegisterTracker(&(*tower));
		// et lancer les hostilites (ouvrir le feu) :
		ChiefManager::GetWeaponManager()->OpenFire(&(*tower), towerKinds[ tower->towerKind ].weaponId);
	}
}

void TowerManager::RefreshFireDir(TSPCell::TwIt& tower)
{
	Vector2i start = tower->rect.GetCenter();
	Vector2i dest	= tower->currentTarget->rect.GetCenter();
	// si la tourelle envoie des projectiles a progression lente, ce n'est pas si simple
	// if faut utiliser un point de tir qui projettera la position de la cible
	// dans le futur quand le projectile interceptera sa course.
	int twwid = towerKinds[ tower->towerKind ].weaponId;
	float prspeed = ChiefManager::GetWeaponManager()->GetProjectileSpeed(twwid);
	if (prspeed > 0.f)  // si les projectiles sont lents
	{
		float targetSpeed = tower->currentTarget->speed;
		float trgtdist = sqrt((float)(dest - start).LenghtSq());
		float time2trgt = trgtdist / prspeed;
		// nouvelle destination de tir, devant l'ennemi ciblé :
		dest = dest + tower->currentTarget->direction * (SPEED_MULTIPLIER * targetSpeed * time2trgt);
	}
	tower->direction = Normalize(dest - start);
	float x = dest.x - start.x;
	float y = dest.y - start.y;
	tower->angle = -1.57f - atan2(y, x);
}
