/**********************************************************************
 * File	: weaponmanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 15/07/2007
 *********************************************************************/

#include "weaponmanager.hpp"
#include "carsmanager.hpp"
#include "terraingenerator.hpp"
#include "hostilesmanager.hpp"
#include "gunner.hpp"
#include "collisionmanager.hpp"

int WeaponManager::instances = 0;

// activer un tir
void WeaponManager::OpenFire(Gunner* a_gunner, int wid)
{
	BOOST_ASSERT(wid <= weapons.GetUbound());

	if (wid < 0 || wid > weapons.GetUbound())
		return;

	// on ne fait évidemment rien si le tireur n'a pas l'arme avec laquelle il veut tirer.
	// rien non plus si ce gunner est deja en train de tirer.
	if (a_gunner->WeaponPresent(wid) && gunners.find(a_gunner) == gunners.end())
	{
		float delay = 0.f;
		float firerate = 1.f / weapons[wid].cyclingRate;
		map< Gunner*, float >::iterator it = oldGunners.find(a_gunner);
		if (it != oldGunners.end())  // si ce tireur existe dans la base des anciens tireurs
		{
			// il ne peut pas tirer avant son prochain fireinterval :
			delay = std::max(0., it->second + firerate - now);  // et ouais mon vieux tu croyais pas t'en tirer si facilement
			oldGunners.erase(it);
		}
		// nouveau tireur:
		gunners[a_gunner] = GunnerInfo(wid, now - firerate + delay);
		// enregistrement de l'arme utilisee et du temps de debut de tir.
		// on vieilli le temps de debut de tir d'un interval de tir pour lancer le premier projectile immediatement
		// (sauf si le dernier tir interrompu est trop recent pour re-tirer, dans ce cas on ajoute un delai.)
	}
}

// arreter un tir
void WeaponManager::CeaseFire(Gunner* a_gunner)
{
	map< Gunner*, GunnerInfo >::iterator it = gunners.find(a_gunner);
	if (it != gunners.end())	// si le tireur existe
	{
		// on doit le supprimer.
		// mais un tireur qui s'arrete ne doit pas etre oublie si facilement !
		// sinon on creer une faille : le delai de repetition peut se contourner
		//										 en repetant plus rapidement l'appui sur la touche de tir.
		// on enregistre donc l'horodatage de l'arret dans une autre base de donnee :
		oldGunners[a_gunner] = it->second.stamp;	// dernier horodatage de tir.
		// maintenant on peut l'enlever :
		gunners.erase(it);
	}
}

void WeaponManager::CreateFireRay(Gunner* pgunr, Vector2f* startIn, Vector2f* dirIn, int direc)
{
	// position de depart pour le tir :
	*startIn = pgunr->GetFireStartCenter();
	const Vector2i offsets = pgunr->GetSideOffsets();

	switch (direc)
	{
		case FD_FRONT:
		{
			*startIn += pgunr->direction * offsets.y;	  // vers l'avant
			*dirIn = pgunr->direction;					  // direction
			break;
		}
		case FD_LEFT:
		{
			*startIn += pgunr->direction.GetOrtho() * offsets.x;
			*dirIn = pgunr->direction.GetOrtho();
			break;
		}
		case FD_RIGHT:
		{
			*startIn -= pgunr->direction.GetOrtho() * offsets.x;
			*dirIn = -pgunr->direction.GetOrtho();
			break;
		}
		case FD_REAR:
		{
			*startIn -= pgunr->direction * offsets.y;
			*dirIn = -pgunr->direction;
			break;
		}
	}
}

void WeaponManager::Update(double gtime)
{
	float current = now = gtime;

	// boucler sur les tireurs actifs
	map< Gunner*, GunnerInfo >::iterator it;
	for (it = gunners.begin(); it != gunners.end(); ++it)
	{
		// determiner si on est plus tard que l'ancien temps de tir + interval de cadence de tir.
		// si oui, on doit tirer une nouvelle salve de projectile(s).
		float fireinterval = 1.f / weapons[it->second.wid].cyclingRate;
		bool fireon = false;
		if (it->second.stamp + fireinterval < current)
		{
			// c'est l'heure d'envoyer un projectile
			fireon = true;
			it->second.stamp += fireinterval;
		}

		// determiner combien de projectiles tire cette arme:
		FIRE_DIRECTION_en ubound = FD_FRONT;
		if (weapons[ it->second.wid ].quadrupleFire)	 // si le gunner se sert d'une arme a tir quadruple
		{
			// on augmente la borne a FD_REAR:
			ubound = FD_REAR;
		}
		// pour chaque projectile que l'on peut tirer simultanement, envoyer un tir si besoin, et/sinon dessiner le flare.
		for (int fireDir = FD_FRONT; fireDir <= ubound; ++fireDir)
		{
			// position de depart et direction pour le tir :
			Vector2f startpos, dir;
			CreateFireRay(it->first, &startpos, &dir, fireDir);

			if (fireon)
			{
				LaunchBullet(startpos, dir, it->first, it->second.wid);
			}

			// modifier un peu la position pour centrer le sprite :
			int flareId = weapons[it->second.wid].flareSprId;
			startpos.x -= sprMngr.GetW(flareId) >> 1;
			startpos.y -= sprMngr.GetH(flareId) >> 1;
			// dessiner les flares a tout moment :
			sprMngr.Draw(weapons[ it->second.wid ].flareSprId, (int)startpos.x, (int)startpos.y);
		}
	}
	// UpdateProjectile :
	//	 - met a jour la progression des projectiles en cours.
	//	 - detecte les collisions avec le monde et les ennemis
	//	 - recompense les tireurs chanceux :)
	//	 - supprime les projectiles mort de la liste de projectiles
	//	 - ajoute des instances d'impacts dans la liste si besoin.
	//	 - endommage les voitures ennemies touchees
	UpdateProjectiles(current - last);
	// dessiner les impacts de la liste circulaire :
	DrawImpacts();
	// lecture du temps
	last = now;
	sprMngr.UpdateWithClientTime(now, UP_TIME_ONLY);
}

void WeaponManager::LaunchBullet(const Vector2f& startpos, const Vector2f& dir, Gunner* a_pgunner, int wid)
{
	// quand on tire quelque chose, on ne sait pas si l'arme utilisée lance des
	// projectiles à impact immédiat (machine gun) ou avec une certaine durée
	// de vie (rocket launcher).
	// donc on va "instancier" des projectiles, et laisser a la fonction
	// UpdateProjectiles la tache de "dés-instanciation" quand ils mourrons.

	// premiere chose, trouver la distance maximum de voyage pour un tir
	// en regardant les collisions avec le terrain
	float t_world = 650.f;	// 650 pixels max de voyage libre
	ChiefManager::GetCollisionsAndProbing()->IntersectRay(&t_world, startpos, dir, false, false);
	// impact avec le terrain precalculé sur toute la durée de vie du projectile.

	// création du projectile et ajout a la liste:
	RegisterProjectile(Projectile(startpos,
								  dir,
								  a_pgunner->angle,
								  t_world,
								  a_pgunner,
								  wid));
	// lancer le son de départ :
	ChiefManager::GetSoundManager()->ReadSound(weapons[wid].sndFire.c_str(), 1, false, startpos);
}

// gestion des projectiles actuellement en l'air.
//	 faire progresser, et impacter.
void WeaponManager::UpdateProjectiles(float delta_time)
{
	// pour tous les projectiles :
	for (int i = 0; i < projectiles.size(); ++i)
	{
		// projectile courant sur lequel on travaille:
		Projectile& proj = projectiles[i];
		bool impactHostile = false;
		Car* carhit;
		if (weapons[ proj.wid ].projectileSpeed <= 0.f)	  // arme a feu immediat (vie du projectile = 0 frame)
		{
			// on lance un rayon, on trouve la collision sur les ennemis la plus proche :
			proj.t = 1e37f;
			impactHostile = ChiefManager::GetHostilesManager()->HostileIntersectRay(&proj.t, &carhit, proj.start, proj.dir);
		}
		else	// arme a projectile progressif (vie du projectile etendue sur plusieurs frames)
		{
			// faire avancer:
			Vector2f projpos = proj.MakeProgress(delta_time, weapons[ proj.wid ].projectileSpeed);
			// dessiner le sprite
			sprMngr.Rotate(weapons[ proj.wid ].projectileSprId, ToDegree(-proj.angle));
			sprMngr.Draw(weapons[ proj.wid ].projectileSprId, (int)projpos.x, (int)projpos.y);
			// detection de l'intersection avec un potentiel ennemi :
			if (proj.t < proj.t_max)
			{
				impactHostile = ChiefManager::GetHostilesManager()->HostileIntersectPoint(&carhit, projpos);
			}
			// clause "sinon" :
			// le projectile est sorti de sa limite d'action
			// ce qui signifie qu'il a touché un obstacle du terrain
			// ou est tombé sur le sol après avoir trop voyagé.
		}
		if (impactHostile)  // si un ennemi est touche.
		{
			// on verifie qu'il soit bien touché avant le terrain:
			if (proj.t < proj.t_max)
			{
				// ok, on a un hit.
				// infliger des dégats à la voiture touchée :
				int damage = weapons[ proj.wid ].damage;
				if (proj.gunner->istower)	// si le tireur est une tourelle, les degats sont attenués.
				{
					damage /= 3;       // diviser par 3
					if (damage <= 0)	 // tout en evitant d'avoir une arme inutile
						damage = 1;
				}
				carhit->InflictDamage(damage);
				if (carhit->health <= 0)
				{
					// récompense si on a tué quelque chose.
					proj.gunner->SignalKills(1);
				}
			}
		}
		if (impactHostile || proj.t > proj.t_max)	 // mort du projectile
		{
			// on joue le sprite d'impact sur le lieu de la mort
			proj.t = std::min(proj.t_max, proj.t);
			// lancer un son pour l'impact:
			if (impactHostile)
				PlayRandomImpactSound(proj.GetPos());
			impacts.push_back(Impact(weapons[ proj.wid ].impactSprId,
											 proj.GetPos(),
											 now));
			// suppression du projectile dans la liste:
			projectiles.pop(i);
			--i;
		}

	}	// fin for ( )
}

void WeaponManager::PlayRandomImpactSound(Vector2i const& pos)
{
	static tools::Random r;
	if (!hitSounds.empty())
	{
		ChiefManager::GetSoundManager()->ReadSound(hitSounds[r.Get(hitSounds.size() - 1)].c_str(), 1, true, pos);
	}
}

void WeaponManager::DrawImpacts(void)
{
	for (int i = 0; i < impacts.size(); ++i)
	{
		if (impacts[i].stamp + 0.3f < now)
		{	// il est mort
			impacts.pop(i);
			--i;
		}
		else
		{
			int posx = (int)impacts[i].pos.x - sprMngr.GetW(impacts[i].sprId) / 2;
			int posy = (int)impacts[i].pos.y - sprMngr.GetH(impacts[i].sprId) / 2;

			sprMngr.Draw(impacts[i].sprId, posx, posy, now - impacts[i].stamp);
		}
	}
}

// rajoute une arme dans la base de donnée des types d'armes
// avec toutes ces informations
int WeaponManager::CreateWeapon(const char* weaponName,
								const char* impactFile,
								const char* flareFile,
								const char* sndFire,
								int cyclingRate,
								int damage,
								int price,
								bool quadrupleFire,
								float projectileSpeed,
								const char* projectileFile)
{
	weapons.Add();

	// chargement des sprites
	int id = sprMngr.LoadSprite(impactFile);	// impact
	weapons.Back().impactSprId = id;
	sprMngr.SetSpeed(id, 10);

	id = sprMngr.LoadSprite(flareFile);	  // flare
	weapons.Back().flareSprId = id;
	sprMngr.SetSpeed(id, 10);

	if (projectileFile != 0)
	{
		id = sprMngr.LoadSprite(projectileFile, SP_ROTATED);	  // projectile
		weapons.Back().projectileSprId = id;
	}
	else
		weapons.Back().projectileSprId = -1;

	// autres informations :
	weapons.Back().name = weaponName;
	weapons.Back().cyclingRate = cyclingRate;
	weapons.Back().damage = damage;
	weapons.Back().price = price;
	weapons.Back().projectileSpeed = projectileSpeed;
	weapons.Back().quadrupleFire = quadrupleFire;

	// copie des chaines de son :
	weapons.Back().sndFire = sndFire;

	// et prépare les samples en eux même :
	if (!ChiefManager::GetSoundManager()->Exist(sndFire))
		ChiefManager::GetSoundManager()->PrepareSoundFile(sndFire, sndFire);

	return weapons.GetUbound();  // retourner l'indice de l'arme a l'user
}

int WeaponManager::FindWeaponID(const char* name)
{
	for (int i = 0; i < weapons.GetElementsNumber(); ++i)
	{
		if (weapons[i].name == name)
			return i;
	}
	return -1;
}
