/**********************************************************************
 * File	: weaponmanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 15/07/2007
 *********************************************************************/

#ifndef WEAPONMANAGER_HPP_INC_19_03_2009___188071817_
#define WEAPONMANAGER_HPP_INC_19_03_2009___188071817_ 1

#include "SecureArray.hpp"
#include "spritemanager.hpp"
#include "timer.hpp"
#include "circularlist.hpp"
#include "boost/array.hpp"
#include "supramanager.hpp"
#include "soundmanager.hpp"

#include <map>
using std::map;
#include <string>
using std::string;

// -----	 nombre de points gagnés par frag :	 -----
#define SCORE_PER_KILL 50
//					  -------------

struct WeaponInfo
{
	int impactSprId;
	int flareSprId;
	string name;

	int cyclingRate;		 // en coups par seconde
	int damage;				 // dommages infliges par impact
	int price;				 // prix de l'arme
	bool quadrupleFire;		 // tir quadruple
	float projectileSpeed;	 // vitesse des projectiles (pixels / seconde)
	                         //  0 ou negatif = impact immediat.
	int projectileSprId;	 // sprite du projectile si l'arme n'est pas a impact immediat.
	string sndFire;			 // nom du son de tir dans le manager de son
};

class Gunner;

class WeaponManager
{
	SpriteManager sprMngr;

	SecureArray< WeaponInfo > weapons;

	struct GunnerInfo
	{
		int wid;
		float stamp;

		GunnerInfo(int id, float st)
				: wid(id),
				  stamp(st) { }
		GunnerInfo(void) { }
	};

	map< Gunner*, GunnerInfo > gunners;	 // tireurs actifs.
	map< Gunner*, float > oldGunners;	 // tireurs qui furent actifs.

	double now;
	float last;

	struct Impact
	{
		int sprId;
		Vector2f pos;
		float stamp;

		Impact(int id, Vector2i p, float t)
				: sprId(id),
				  pos(p),
				  stamp(t) { }
		Impact(void) { }
	};
	CircularList< Impact > impacts;

	struct Projectile
	{
		Vector2f start;
		Vector2f dir;
		float angle;
		float t;
		float t_max;
		Gunner* gunner;
		int wid;

		Vector2f GetPos(void)
		{
			return start + dir * t;
		}
		Vector2f MakeProgress(float delta_time, float speed)
		{
			t += delta_time * speed;
			return GetPos();
		}

		Projectile(void) { }
		Projectile(const Vector2f& a_st, const Vector2f& a_dr, float a_angle, float maxdist, Gunner* pgunner, int a_wid)
				: start(a_st),
				  dir(a_dr),
				  angle(a_angle),
				  t(0),
				  t_max(maxdist),
				  gunner(pgunner),
				  wid(a_wid) { }
	};
	CircularList< Projectile > projectiles;

	// enumeration des directions de tir:
	enum FIRE_DIRECTION_en
	{
		FD_FRONT,
		FD_LEFT,
		FD_RIGHT,
		FD_REAR
	};

	std::vector< std::string > hitSounds;

	// -- methodes privees --

	void LaunchBullet(const Vector2f& startpos, const Vector2f& dir, Gunner* gunner, int wid);
	void PushImpact(int sprId, Vector2i pos);
	void DrawImpacts(void);
	void CreateFireRay(Gunner* gunr, Vector2f* startIn, Vector2f* dirIn, int direc = FD_FRONT);

	void RegisterProjectile(const Projectile& a_proj)
	{
		projectiles.push_back(a_proj);
	}
	void UpdateProjectiles(float delta_time);
	
	void PlayRandomImpactSound(Vector2i const& pos);

	static int instances;

public:

	struct TooManyInstances {};

	WeaponManager(void)
			: now(0.),
			  last(0.f)
	{
		if (++instances != 1)
			throw TooManyInstances();
		weapons.ReserveExtra(5);
		impacts.resize(100);
		projectiles.resize(1000);
	}
	~WeaponManager()
	{
		--instances;
	}

	int GetCyclingRate(int wid)
	{
		return weapons[wid].cyclingRate;
	}

	int GetPrice(int wid)
	{
		return weapons[wid].price;
	}

	float GetProjectileSpeed(int wid)
	{
		return weapons[wid].projectileSpeed;
	}

	void OpenFire(Gunner* a_gunner, int wid);
	void CeaseFire(Gunner* a_gunner);

	void Update(double global_time);

	int FindWeaponID(const char* name);

	int CreateWeapon(const char* weaponName,
					 const char* impactFile,
					 const char* flareFile,
					 const char* sndFire,
					 int cyclingRate,
					 int damage,
					 int price = 0,
					 bool quadrupleFire = false,
					 float projectileSpeed = 0.f,
					 const char* projectileFile = 0);

	template< size_t N >
	void SetCarImpactSounds(boost::array< char const*, N > const& sounds);
};

// necessary inline implementation:
template< size_t N >
void WeaponManager::SetCarImpactSounds(boost::array< char const*, N > const& sounds)
{
	hitSounds.reserve(N);
	for (size_t i = 0; i < N; ++i)
	{
		hitSounds.push_back(sounds[i]);
		ChiefManager::GetSoundManager()->PrepareSoundFile(sounds[i], sounds[i]);  // ici le chemin fera identifiant aussi, ca ira.
	}
}

#endif //WEAPONMANAGER_HPP_INC_19_03_2009___188071817_
