/**********************************************************************
 * File  : freelist.hpp
 * Author: Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 31/07/2007
 *********************************************************************/

#ifndef FREELIST_HPP_INC_19_03_2009__102520996_
#define FREELIST_HPP_INC_19_03_2009__102520996_ 1

//! should be somewhat equivalent to something like : std::list< T, boost::fast_allocator >

//! a free list is a container allocating space by removing the empty slots instead of allocating them.
//! this can be a lot faster than a classic list.

//! profiling is giving the following results :
//!
//!   - pushing is 5 times faster than STL std::list as soon as nodes are over size 10.
//!      6 times faster with size 100, 7 times with 1000.
//!   - iterating is nearly the same performance
//!   - erasing is 1 to 400 times faster

//! WARNING NOTE : beware, freelist is not thread safe !

#include <cstddef>
#include <algorithm>

#ifndef BOOST_ASSERT
# define BOOST_ASSERT(x)
#endif

namespace tools
{

template< typename T >
class freelist
{
	// -- private --

	template< typename TT >
	struct node
	{
		TT* pool;

		int last;
		const int size;

		node* next;

		bool push_back(const TT& val)
		{
			// check if we don't overrun the capacity :
			if (last == size)
			{
				// then signal impossible to push:
				return false;
			}
			pool[last] = val;
			++last;
			return true;
		}

		node(int psize)
		: last(0),
		  size(psize),
		  next(0)
		{
			pool = new T[size];
		}
		~node()
		{
			delete [] pool;
		}
	};

	template< typename TI >
	struct internal_iterator_base
	{
		internal_iterator_base(node< TI >* begin)
		: cur(begin),
		  index(0)
		{	}

		internal_iterator_base(void)
		: cur(0),
		  index(0)
		{  }

		internal_iterator_base& operator ++ (void)
		{
			++index;
			if (index >= cur->last)  // we must jump between nodes
			{
				cur = cur->next;
				index = 0;
			}
			return *this;
		}

		bool operator != (const internal_iterator_base& op) const
		{
			return cur != op.cur || index != op.index;
		}

		// two const operations in the base
		const TI& operator * (void) const
		{
			return cur->pool[index];
		}

		const TI* operator -> (void) const
		{
			return &cur->pool[index];
		}

	protected:

		node< TI >* cur;
		int index;
	};

	template< typename TI >
	struct internal_iterator : internal_iterator_base<TI>
	{
		internal_iterator(node< TI >* begin) : internal_iterator_base<TI>(begin) {}
		internal_iterator(void) : internal_iterator_base<TI>() {}

		// two supplementary non-const operations
		TI& operator * (void)
		{
			return this->cur->pool[this->index];
		}

		TI* operator -> (void)
		{
			return &this->cur->pool[this->index];
		}
	};

	template< typename TI >
	struct const_internal_iterator : internal_iterator_base<TI>
	{
		const_internal_iterator(node< TI >* begin) : internal_iterator_base<TI>(begin) {}
		const_internal_iterator(void) : internal_iterator_base<TI>() {}
		const_internal_iterator(internal_iterator<TI> const & ii) : internal_iterator_base<TI>(ii) {}
	};

	node< T >* first;
	node< T >* last;

	int pool_size;
	size_t num_elems;   // tracker for size() function

	freelist& Copy(freelist<T> const & right);  // private to unify copy constructor and operator = code
	inline void InitEmpty(int pool_size);

public:

	typedef internal_iterator< T > iterator;
	typedef const_internal_iterator< T > const_iterator;

	iterator begin(void);
	iterator end(void);
	const_iterator begin(void) const;
	const_iterator end(void) const;

	void push_back(const T& val);
	size_t size(void) const { return num_elems; }   //!< STL size()-like but guaranteed O(1)

	T& back(void);  //!< get last element
	const T& back(void) const;
	T& front(void);  //!< get first element
	const T& front(void) const;

	void clear(void);
	bool empty() const { return size() == 0; }

	void swap(freelist<T>& right);  //!< exchange the content and pool policies in O(1)

	~freelist();
	freelist(int free_pool_size = 100);
	freelist(freelist<T> const & right);  // copy construction

	freelist& operator = (freelist<T> const & right);  // assignation
};

//! provide a template struct to make an STL compatible type.
template< size_t Size >  // in-template parameterizable pool size.
struct MetaFreeList
{
	template< typename T, typename Alloc >  // match STL number of arguments (Alloc is ignored here)
	struct STLCompliantT : tools::freelist< T >
	{
		STLCompliantT() : tools::freelist< T >(Size) {}

		template< size_t Size2 >
		STLCompliantT(typename MetaFreeList< Size2 >::template STLCompliantT< T, Alloc > const & right)
		: tools::freelist< T >(right)
		{}
	};
};

// === definitions of freelist class ===

// constructor :
template< typename T >
freelist< T >::freelist(int free_pool_size)
{
	InitEmpty(free_pool_size);
}

// copy construction (assigns pool policy of right operand)
template< typename T >
freelist< T >::freelist(freelist<T> const & right)
{
	InitEmpty(right.pool_size);
	Copy(right);
}

// assignation
template< typename T >
freelist<T>& freelist< T >::operator = (freelist<T> const & right)
{
	return Copy(right);
}

template< typename T >
void freelist< T >::InitEmpty(int a_pool_size)  // private
{
	first = 0;
	last = 0;
	pool_size = (a_pool_size < 1) ? 1 : a_pool_size;  // it is forbidden, and stupid, to specify less-than-1 pool-size.
	num_elems = 0;
}

//! class internal copy function. Warn: does NOT copy pool-size policy. O(N) complexity.
template< typename T >
freelist<T>& freelist< T >::Copy(freelist<T> const & right)  // private
{
	if (this != &right)
	{
		clear();  // empty this object

		// iterate through the right freelist pushing in the left (this) one.
		typename freelist<T>::const_iterator it = right.begin();
		while (it != right.end())
		{
			push_back(*it);
			++it;
		}
	}
	return (*this);
}

// destructor :
template< typename T >
freelist< T >::~freelist()
{
	clear();
}

template< typename T >
void freelist< T >::clear(void)
{
	// if there is something. let's deallocate the total
	if (first)
	{
		node<T>* cur = first;
		node<T>* next;
		do
		{
			next = cur->next;
			delete cur;
			cur = next;
		} while (cur != 0);
		first = 0;
		num_elems = 0;
	}
}

template< typename T >
void freelist< T >::push_back(const T& val)
{
	if (first == 0)  // empty list
	{
		first = last = new node< T >(pool_size);
	}
	bool ok = last->push_back(val);
	if (!ok)  // need to allocate a new node
	{
		node<T>* newnode = new node< T >(pool_size);
		last->next = newnode;
		last = newnode;
		last->push_back(val);  // if this one is false, ..well its impossible.
	}
	++num_elems;
}

template< typename T >
T& freelist< T >::back(void)  // get last element
{
	return last->pool[last->last - 1];
}

template< typename T >
const T& freelist< T >::back(void) const
{
	return last->pool[last->last - 1];
}

template< typename T >
T& freelist< T >::front(void)  // get first element
{
	BOOST_ASSERT(first);
	return first->pool[0];
}

template< typename T >
const T& freelist< T >::front(void) const
{
	BOOST_ASSERT(first);
	return first->pool[0];
}

template< typename T >
typename freelist<T>::iterator freelist< T >::begin(void)
{
	return iterator(first);
}

template< typename T >
typename freelist<T>::iterator freelist< T >::end(void)
{
	return iterator(0);
}

template< typename T >
typename freelist<T>::const_iterator freelist< T >::begin(void) const
{
	return const_iterator(first);
}

template< typename T >
typename freelist<T>::const_iterator freelist< T >::end(void) const
{
	return const_iterator(0);
}

template< typename T >
void freelist< T >::swap(freelist<T>& right)
{
	std::swap(first,     right.first);
	std::swap(last,      right.last);
	std::swap(pool_size, right.pool_size);
	std::swap(num_elems, right.num_elems);
}

}

#endif //FREELIST_HPP_INC_19_03_2009__102520996_
