/**********************************************************************
 * File	: fpscomputer.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *			  mailto:lightness1024@free.fr
 * Date	: 25/08/2008
 *********************************************************************/

#ifndef PERFORMANCEPROFILER_HPP_INC_19_03_2009___220193600_
#define PERFORMANCEPROFILER_HPP_INC_19_03_2009___220193600_ 1

#include "freelist.hpp"
#include <map>
#include <string>
#include <algorithm>
#include <iostream>
#include "smart_ptr.hpp"
#include <limits>

#undef min
#undef max

namespace tools
{

class PerformanceProfiler
{
public:

	struct Stats
	{
		float avg;
		float min;
		float max;
		float last;
		size_t numRec;

		Stats(void)
			:	avg(0.f),
				min(std::numeric_limits<float>::max()),
				max(-1.f),
				last(0.f),
				numRec(0)
		{ }
		// a Stat is ordered from the average :
		bool operator < (const Stats& rValue) const
		{
			return avg < rValue.avg;
		}
		bool operator == (const Stats& rValue) const
		{
			return avg == rValue.avg;
		}
	};

private:

	struct TimeStamp
	{
		float timeline;
		float duration;

		TimeStamp(float tl, float du) : timeline(tl), duration(du) {}
		TimeStamp() {}
	};

	struct TimeStampsList
	{
		typedef tools::freelist< TimeStamp > StampsList;

		smart_ptr< StampsList > stamp_list;  // pointers make object fast to copy, and riskless regarding data
		bool began;

		Stats GetStats(void) const
		{
			Stats res;

			StampsList::iterator it;
			for (it = stamp_list->begin(); it != stamp_list->end(); ++it)
			{
				res.avg += it->duration;    // avg
				if (res.min > it->duration)
					res.min = it->duration; // min
				if (res.max < it->duration)
					res.max = it->duration; // max
			}
			res.numRec = stamp_list->size();
			res.avg /= stamp_list->size();
			res.last = stamp_list->back().duration;
			return res;
		}

		TimeStampsList(void)
			: stamp_list(new StampsList),
			  began(false)
		{ }
	};

	typedef std::map< std::string, TimeStampsList > MapOfProfiles;

	MapOfProfiles profiles;

public:

	void BeginProfile(const std::string& profile_name, float start_time)
	{
		// chercher le profil dans la map:
		MapOfProfiles::iterator profile = profiles.find(profile_name);

		// si ce profil existe et qu'il y a dj eu un Begin sans End sur ce profil
		if (profile != profiles.end() && profile->second.began)
		{
			// on ne crer pas de nouveau stamp mais on met a jour l'horodatage de dpart
			profile->second.stamp_list->back().timeline = start_time;
		}
		else
		{
			// la prochaine action crer un nouveau profil dans la map si "profile_name" n'existe pas
			// elle crer aussi un nouveau stamp dans la liste du profil, et stocke le temps de dpart
			profiles[profile_name].stamp_list->push_back(TimeStamp(start_time, 0.f));  
			// et on passe la liste en mode capture
			profiles[profile_name].began = true;
		}
	}

	void EndProfile(const std::string& profile_name, float end_time)
	{
		// chercher le profil dans la map:
		MapOfProfiles::iterator profile = profiles.find(profile_name);
		
		// si existe pas, on n'a pas fait begin, donc End n'a rien  faire
		// de mme si le profile existe mais aucune capture n'a t dmarre
		if (profile == profiles.end() || !profile->second.began)
			return;

		// mise  jour de la duration :
		profile->second.stamp_list->back().duration = end_time - profile->second.stamp_list->back().timeline;  // temps actuel moins temps de dpart
		profile->second.began = false;  // on marque la capture comme arrte pour cette tranche de temps
	}

	Stats GetStats(const std::string& profile_name) const
	{
		MapOfProfiles::const_iterator profile = profiles.find(profile_name);
		if (profile == profiles.end())
			return Stats();
		else
			return profile->second.GetStats();
	}

	void Flush(void)
	{
		profiles.clear();
	}

	// systme d'affichage
	friend std::ostream& operator << (std::ostream& os, const PerformanceProfiler& perf);
};

// make a structure "Stats" streamable
inline std::ostream& operator << (std::ostream& os, const PerformanceProfiler::Stats& stats)
{
	os << "avg: " << stats.avg << "s. min: " << stats.min << "s. max: " << stats.max << "s. num records: " << stats.numRec;
	return os;
}

// foncteur pour for_each
struct PrintPerfProfileRecord
{
	std::ostream& m_os;
	PrintPerfProfileRecord(std::ostream& os) : m_os(os) {}

	void operator () (std::map< PerformanceProfiler::Stats, std::string >::value_type pair)
	{
		m_os << "profile: " << pair.second << ", " << pair.first << std::endl;
	}
};

inline std::ostream& operator << (std::ostream& os, const PerformanceProfiler& perf)
{
	std::map< PerformanceProfiler::Stats, std::string > reverseMap;
	PerformanceProfiler::MapOfProfiles::const_iterator it;

	// remplissage de la map inverse pour trier par ordre de dure
	for (it = perf.profiles.begin(); it != perf.profiles.end(); ++it)
		reverseMap[it->second.GetStats()] = it->first;
	
	PrintPerfProfileRecord printer(os);
	std::for_each(reverseMap.begin(), reverseMap.end(), printer);
	return os;
}

}  // end namespace tools


#endif //PERFORMANCEPROFILER_HPP_INC_19_03_2009___220193600_
