/**********************************************************************
 * File	: smart_ptr.hpp
 * Author: Oddou Vivien - a.k.a. Lightness1024!
 *			  mailto:vivien@isit.or.jp
 * Date	: 17/05/2007
 *********************************************************************/

/*
 *	 -- WHAT IS A SMART POINTER ? --
 *
 * this file declares and defines a class of smart pointer.
 * a smart pointer bechaves like a C-pointer. Except that when it goes out of scope,
 * or is destroyed, it triggers the destruction of the pointed object.
 * this smart pointer also has the convenient feature of reference counting.
 * if many smart pointers are copied, and therefore points to the same object,
 * then the destruction of one of them will not imply the destruction of the pointed object,
 * until there is only one pointer left.
 * A very good document for state-of-the-art programming of smart pointers can be found here:
 *		http://www.awprofessional.com/articles/article.asp?p=31529&seqNum=1&rl=1
 */


#ifndef SMART_PTR_HPP_INC_19_03_2009___196955462_
#define SMART_PTR_HPP_INC_19_03_2009___196955462_ 1

template< typename T >
class smart_ptr
{
protected:

	T* m_ptr;
	int* refcount;			// shared reference counter

	// note to implementor : beware not to call any Acquire-style function without having Released before.
	void Acquire(const smart_ptr& a_right)
	{
		m_ptr = a_right.m_ptr;
		refcount = a_right.refcount;
		++*refcount;
	}

	void Release(void)
	{
		--*refcount;
		if (*refcount == 0)
		{
			if (m_ptr)
				delete m_ptr;
			delete refcount;
			m_ptr = 0;
			refcount = 0;
		}
	}

public:

	// default construction
	smart_ptr(void)
			: m_ptr(0),
			  refcount(new int(1))
	{}

	// creation from raw pointer :
	explicit smart_ptr(T* a_object) : m_ptr(a_object),
									  refcount(new int(1))
	{}

	// creation from another smart_ptr :
	smart_ptr(const smart_ptr& a_2cpy)
	{
		Acquire(a_2cpy);
	}

	// deletion :
	~smart_ptr()
	{
		Release();
	}

	// copy from = operator
	const T& operator = (const smart_ptr& a_right)
	{
		Release();
		Acquire(a_right);
		return *m_ptr;
	}

	// access to data :
	T& operator * (void) const
	{
		return *m_ptr;
	}
	T* operator -> (void) const
	{
		return m_ptr;
	}

	// raw access to pointer (warning, this is a leak)
	T* GetPtr(void) const
	{
		return m_ptr;
	}
	
	int GetCount(void)
	{
		return refcount ? *refcount : 0;
	}
};

#endif //SMART_PTR_HPP_INC_19_03_2009___196955462_
